import json
import numbers
import time
import numpy as np
from collections import defaultdict
from itertools import product
from numba import jit
from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.utils.multiclass import type_of_target
from sklearn.utils.validation import check_X_y, check_array, check_is_fitted
from sklearn.utils import check_random_state
from joblib import Parallel, delayed
from sklearn.base import clone
from .util import convert_numpy


_TREE_LEAF = -1
_TREE_UNDEFINED = -2

LEFT = 0
LEFT_INTERSECT = 1
RIGHT_INTERSECT = 2
RIGHT = 3

NOGIL = True


class Node:
    """Base class for decision tree nodes, also functions as leaf."""

    def __init__(self, feature, left_child, right_child, value):
        self.feature = feature
        self.left_child = left_child
        self.right_child = right_child
        self.value = value

    def predict(self, _):
        assert self.left_child == _TREE_LEAF
        assert self.right_child == _TREE_LEAF

        return self.value

    def pretty_print(self, depth=0):
        indentation = depth * "  "
        if isinstance(self.value, np.ndarray):
            return f"{indentation}return [{self.value[0]:.3f}, {self.value[1]:.3f}]"
        else:
            return f"{indentation}return {self.value:.3f}"

    def to_json(self):
        if isinstance(self.value, np.ndarray):
            return {
                "value": [self.value[0], self.value[1]],
            }
        else:
            return {
                "value": self.value,
            }

    def to_xgboost_json(self, node_id, depth):
        if isinstance(self.value, np.ndarray):
            # Return leaf value in range [-1, 1]
            return {"nodeid": node_id, "leaf": self.value[1] * 2 - 1}, node_id
        else:
            return {"nodeid": node_id, "leaf": self.value}, node_id

    def is_leaf(self):
        return self.left_child == _TREE_LEAF and self.right_child == _TREE_LEAF

    def prune(self, _):
        return self


class NumericalNode(Node):
    """
    Decision tree node for numerical decision (threshold).
    """

    def __init__(self, feature, threshold, left_child, right_child, value):
        super().__init__(feature, left_child, right_child, value)
        self.threshold = threshold

    def predict(self, sample):
        """
        Predict the class label of the given sample. Follow the left subtree
        if the sample's value is lower or equal to the threshold, else follow
        the right sub tree.
        """
        comparison = sample[self.feature] <= self.threshold
        if comparison:
            return self.left_child.predict(sample)
        else:
            return self.right_child.predict(sample)

    def pretty_print(self, depth=0):
        indentation = depth * "  "
        return f"""{indentation}if x{self.feature} <= {self.threshold}:
{self.left_child.pretty_print(depth + 1)}
{indentation}else:
{self.right_child.pretty_print(depth + 1)}"""

    def to_json(self):
        return {
            "feature": self.feature,
            "threshold": self.threshold,
            "left_child": self.left_child.to_json(),
            "right_child": self.right_child.to_json(),
        }

    def to_xgboost_json(self, node_id, depth):
        left_id = node_id + 1
        left_dict, new_node_id = self.left_child.to_xgboost_json(left_id, depth + 1)

        right_id = new_node_id + 1
        right_dict, new_node_id = self.right_child.to_xgboost_json(right_id, depth + 1)

        return (
            {
                "nodeid": node_id,
                "depth": depth,
                "split": self.feature,
                "split_condition": self.threshold,
                "yes": left_id,
                "no": right_id,
                "missing": left_id,
                "children": [left_dict, right_dict],
            },
            new_node_id,
        )

    def prune(self, bounds=defaultdict(lambda: [-np.inf, np.inf])):
        old_high = bounds[self.feature][1]
        bounds[self.feature][1] = self.threshold

        self.left_child = self.left_child.prune(bounds)

        bounds[self.feature][1] = old_high
        old_low = bounds[self.feature][0]
        bounds[self.feature][0] = self.threshold

        self.right_child = self.right_child.prune(bounds)

        bounds[self.feature][0] = old_low

        if self.threshold >= bounds[self.feature][1] or self.threshold == np.inf:
            # If no sample can reach this node's right side
            return self.left_child
        elif self.threshold <= bounds[self.feature][0] or self.threshold == -np.inf:
            # If no sample can reach this node's left side
            return self.right_child
        elif (
            self.left_child.is_leaf()
            and self.right_child.is_leaf()
            and self.left_child.value[1] == self.right_child.value[1]
        ):
            # If both children are leaves and they predict the same value
            return self.left_child
        else:
            return self


def _attack_model_to_tuples(attack_model):
    new_attack_model = []
    for attack_mode in attack_model:
        if attack_mode == "":
            new_attack_model.append((0, 0))
        elif attack_mode == ">":
            new_attack_model.append((0, 10e9))
        elif attack_mode == "<":
            new_attack_model.append((10e9, 0))
        elif attack_mode == "<>":
            new_attack_model.append((10e9, 10e9))
        elif isinstance(attack_mode, numbers.Number):
            new_attack_model.append((attack_mode, attack_mode))
        elif isinstance(attack_mode, tuple) and len(attack_mode) == 2:
            new_attack_model.append(attack_mode)
        else:
            raise Exception("Unknown attack model spec:", attack_mode)
    return new_attack_model


@jit(nopython=True, nogil=NOGIL)
def _scan_numerical_feature_fast(
    samples,
    y,
    dec,
    inc,
    left_bound,
    right_bound,
    chen_heuristic,
    one_adversarial_class,
):
    sort_order = samples.argsort()
    sorted_labels = y[sort_order]
    sample_queue = samples[sort_order]
    dec_queue = sample_queue - dec
    inc_queue = sample_queue + inc

    # Initialize sample counters
    l_0 = l_1 = li_0 = li_1 = ri_0 = ri_1 = 0
    label_counts = np.bincount(y)
    r_0 = label_counts[0]
    r_1 = label_counts[1]

    # Initialize queue values and indices
    sample_i = dec_i = inc_i = 0
    sample_val = sample_queue[0]
    dec_val = dec_queue[0]
    inc_val = inc_queue[0]

    best_score = 10e9
    best_split = None
    adv_gini = None
    while True:
        smallest_val = min(sample_val, dec_val, inc_val)

        # Find the current point and label from the queue with smallest value.
        # Also update the sample counters
        if sample_val == smallest_val:
            point = sample_val
            label = sorted_labels[sample_i]

            if label == 0:
                if one_adversarial_class:
                    r_0 -= 1
                    l_0 += 1
                else:
                    ri_0 -= 1
                    li_0 += 1
            else:
                ri_1 -= 1
                li_1 += 1

            # Update sample_val and i to the values belonging to the next
            # sample in queue. If we reached the end of the queue then store
            # a high number to make sure the sample_queue does not get picked
            if sample_i < sample_queue.shape[0] - 1:
                sample_i += 1
                sample_val = sample_queue[sample_i]
            else:
                sample_val = 10e9
        elif dec_val == smallest_val:
            point = dec_val
            label = sorted_labels[dec_i]

            if label == 0:
                if not one_adversarial_class:
                    r_0 -= 1
                    ri_0 += 1
            else:
                r_1 -= 1
                ri_1 += 1

            # Update dec_val and i to the values belonging to the next
            # sample in queue. If we reached the end of the queue then store
            # a high number to make sure the dec_queue does not get picked
            if dec_i < dec_queue.shape[0] - 1:
                dec_i += 1
                dec_val = dec_queue[dec_i]
            else:
                dec_val = 10e9
        else:
            point = inc_val
            label = sorted_labels[inc_i]

            if label == 0:
                if not one_adversarial_class:
                    li_0 -= 1
                    l_0 += 1
            else:
                li_1 -= 1
                l_1 += 1

            # Update inc_val and i to the values belonging to the next
            # sample in queue. If we reached the end of the queue then store
            # a high number to make sure the inc_queue does not get picked
            if inc_i < inc_queue.shape[0] - 1:
                inc_i += 1
                inc_val = inc_queue[inc_i]
            else:
                inc_val = 10e9

        if point >= right_bound:
            break

        # If the next point is not the same as this one
        next_point = min(sample_val, dec_val, inc_val)
        if next_point != point:
            if one_adversarial_class:
                if chen_heuristic:
                    adv_gini, _ = chen_adversarial_gini_gain_one_class(
                        l_0, l_1, r_0, r_1, li_1, ri_1
                    )
                else:
                    adv_gini, _ = adversarial_gini_gain_one_class(
                        l_0, l_1, r_0, r_1, li_1 + ri_1
                    )
            else:
                if chen_heuristic:
                    adv_gini, _, __ = chen_adversarial_gini_gain_two_class(
                        l_0, l_1, li_0, li_1, ri_0, ri_1, r_0, r_1
                    )
                else:
                    adv_gini, _, __ = adversarial_gini_gain_two_class(
                        l_0, l_1, li_0, li_1, ri_0, ri_1, r_0, r_1
                    )

            # Maximize the margin of the split
            split = (point + next_point) * 0.5

            if (
                adv_gini is not None
                and adv_gini < best_score
                and split > left_bound
                and split < right_bound
            ):
                best_score = adv_gini
                best_split = split

    return best_score, best_split


# @jit(nopython=True, nogil=NOGIL)
def _scan_numerical_feature_fast_regression(
    samples,
    y,
    dec,
    inc,
    left_bound,
    right_bound,
    chen_heuristic,
):
    unique_samples = np.unique(samples)

    if dec == 0 and inc == 0:
        thresholds = np.sort(unique_samples)
    else:
        thresholds = np.sort(
            np.unique(np.concatenate((unique_samples - dec, unique_samples + inc)))
        )

    samples_inc = samples + inc
    samples_dec = samples - dec

    best_score = 10e9
    best_split = None
    adv_sse = None
    for point, next_point in zip(thresholds[:-1], thresholds[1:]):

        if point >= right_bound:
            break

        y_left = y[samples_inc <= point]
        y_right = y[samples_dec > point]

        if chen_heuristic:
            y_left_intersect = y[(samples <= point) & (samples_inc > point)]
            y_right_intersect = y[(samples > point) & (samples_dec <= point)]

            adv_sse, _ = chen_adversarial_sum_absolute_errors(
                y_left,
                y_left_intersect,
                y_right_intersect,
                y_right,
            )
        else:
            y_intersect = y[~((samples_inc <= point) | (samples_dec > point))]

            adv_sse, _ = adversarial_sum_absolute_errors(
                y_left,
                y_right,
                y_intersect,
            )

        # Maximize the margin of the split
        split = (point + next_point) * 0.5

        if (
            adv_sse is not None
            and adv_sse < best_score
            and split > left_bound
            and split < right_bound
        ):
            best_score = adv_sse
            best_split = split

    return best_score, best_split


@jit(nopython=True, nogil=NOGIL)
def chen_adversarial_gini_gain_one_class(l_0, l_1, r_0, r_1, li_1, ri_1):
    i_1 = li_1 + ri_1

    s1 = weighted_gini(l_0, l_1 + li_1, r_0, r_1 + ri_1)
    s2 = weighted_gini(l_0, l_1, r_0, r_1 + i_1)
    s3 = weighted_gini(l_0, l_1 + i_1, r_0, r_1)
    s4 = weighted_gini(l_0, l_1 + ri_1, r_0, r_1 + li_1)

    worst_case = max(s1, s2, s3, s4)

    # Return the worst found weighted Gini impurity, the number of class 1
    # samples that move to the left and the number of class 0 samples that
    # move to the left
    if s1 == worst_case:
        return s1, li_1

    if s2 == worst_case:
        return s2, 0

    if s3 == worst_case:
        return s3, i_1

    if s4 == worst_case:
        return s4, ri_1


@jit(nopython=True, nogil=NOGIL)
def chen_adversarial_gini_gain_two_class(l_0, l_1, li_0, li_1, ri_0, ri_1, r_0, r_1):
    i_0 = li_0 + ri_0
    i_1 = li_1 + ri_1

    s1 = weighted_gini(l_0 + li_0, l_1 + li_1, r_0 + ri_0, r_1 + ri_1)
    s2 = weighted_gini(l_0, l_1, r_0 + i_0, r_1 + i_1)
    s3 = weighted_gini(l_0 + i_0, l_1 + i_1, r_0, r_1)
    s4 = weighted_gini(l_0 + ri_0, l_1 + ri_1, r_0 + li_0, r_1 + li_1)

    worst_case = max(s1, s2, s3, s4)

    # Return the worst found weighted Gini impurity, the number of class 1
    # samples that move to the left and the number of class 0 samples that
    # move to the left
    if s1 == worst_case:
        return s1, li_1, li_0

    if s2 == worst_case:
        return s2, 0, 0

    if s3 == worst_case:
        return s3, i_1, i_0

    if s4 == worst_case:
        return s4, ri_1, ri_0


@jit(nopython=True, nogil=NOGIL)
def adversarial_gini_gain_one_class(l_0, l_1, r_0, r_1, i_1):
    # Fast implementation of the adversarial Gini gain, it finds the
    # analytical maximum and rounds to the nearest two ints, then returns
    # the highest of those two. x is limited by the range [0, i_1].
    x = max(min((l_0 * r_1 + l_0 * i_1 - l_1 * r_0) / (l_0 + r_0), i_1), 0)

    x_floor = int(np.floor(x))
    x_ceil = int(np.ceil(x))
    adv_gini_floor = weighted_gini(l_0, l_1 + x_floor, r_0, r_1 + i_1 - x_floor)
    adv_gini_ceil = weighted_gini(l_0, l_1 + x_ceil, r_0, r_1 + i_1 - x_ceil)
    if adv_gini_floor > adv_gini_ceil:
        return adv_gini_floor, x_floor
    else:
        return adv_gini_ceil, x_ceil


@jit(nopython=True, nogil=NOGIL)
def adversarial_gini_gain_two_class(l_0, l_1, li_0, li_1, ri_0, ri_1, r_0, r_1):
    i_0 = li_0 + ri_0
    i_1 = li_1 + ri_1

    if i_0 == 0 and i_1 == 0:
        return weighted_gini(l_0, l_1, r_0, r_1), 0, 0

    if l_1 + r_1 + i_1 == 0:
        return (
            weighted_gini(l_0 + li_0, l_1 + li_1, r_0 + i_0 - li_0, r_1 + i_1 - li_1),
            li_1,
            li_0,
        )

    # Compute these before since we use their values multiple times
    x_coef = (l_0 + r_0 + i_0) / (l_1 + r_1 + i_1)
    intercept = (l_1 * r_0 - l_0 * r_1 - l_0 * i_1 + l_1 * i_0) / (l_1 + r_1 + i_1)
    denominator = x_coef ** 2 + 1

    # In the paper we refer to m1, m0 here they are li_1 and li_0
    x_prime = round((li_1 + x_coef * (li_0 - intercept)) / denominator)
    y_prime = round((x_coef * (li_1 + x_coef * li_0) + intercept) / denominator)

    # Unfortunately the best solution often lies outside our region of interest
    # (x in [0, i_1] and y in [0, i_0]) so we have to check for corner cases
    if x_prime < 0 and y_prime > i_0:
        # If the point (x', y') is out the top-left corner of our region of
        # interest then the line does not pass the region, we use the best
        # point which is that corner
        x_prime = 0
        y_prime = i_0
    elif x_prime < 0:
        # If x' is smaller than 0 we try the closest point on the solution line
        # in the region x \in [0, i_1] which is x = 0
        x_prime = 0
        y_prime = (
            l_1 * r_0 - l_0 * r_1 - l_0 * i_1 + l_1 * i_0 + (l_0 + r_0 + i_0) * x_prime
        ) / (l_1 + r_1 + i_1)
        if y_prime > i_0:
            # If y is still not in the region than the line is completely
            # outside of the region
            x_prime = 0
            y_prime = i_0
    elif x_prime > i_1 and y_prime < 0:
        # If the point (x', y') is out the bottom-right corner of our region of
        # interest then the line does not pass the region, we use the best
        # point which is that corner
        x_prime = i_1
        y_prime = 0
    elif x_prime > i_1:
        # If x' is larger than i_10 we try the closest point on the solution
        # line in the region x \in [0, i_1] which is x = i_1
        x_prime = i_1
        y_prime = (
            l_1 * r_0 - l_0 * r_1 - l_0 * i_1 + l_1 * i_0 + (l_0 + r_0 + i_0) * x_prime
        ) / (l_1 + r_1 + i_1)
        if y_prime < 0:
            # If y is still not in the region than the line is completely
            # outside of the region
            x_prime = i_1
            y_prime = 0
    elif y_prime < 0:
        # If y' is smaller than 0 we try the closest point on the solution line
        # in the region y \in [0, i_1] which is y = 0
        y_prime = 0
        x_prime = (
            l_0 * r_1 + l_0 * i_1 - l_1 * r_0 - l_1 * i_0 + (l_1 + r_1 + i_1) * y_prime
        ) / (l_0 + r_0 + i_0)
        if x_prime > i_1:
            x_prime = i_1
            y_prime = 0
    elif y_prime > i_0:
        # If y' is smaller than 0 we try the closest point on the solution line
        # in the region y \in [0, i_1] which is y = 0
        y_prime = i_0
        x_prime = (
            l_0 * r_1 + l_0 * i_1 - l_1 * r_0 - l_1 * i_0 + (l_1 + r_1 + i_1) * y_prime
        ) / (l_0 + r_0 + i_0)
        if x_prime < 0:
            x_prime = 0
            y_prime = i_0

    x_prime = int(round(x_prime))
    y_prime = int(round(y_prime))

    assert x_prime >= 0 and x_prime <= i_1
    assert y_prime >= 0 and y_prime <= i_0

    # Return the gini gain given the rounded x and y prime
    return (
        weighted_gini(
            l_0 + y_prime, l_1 + x_prime, r_0 + i_0 - y_prime, r_1 + i_1 - x_prime
        ),
        x_prime,
        y_prime,
    )


@jit(nopython=True, nogil=NOGIL)
def gini_impurity(i_0, i_1):
    if i_0 + i_1 == 0:
        return 1.0

    ratio = i_0 / (i_0 + i_1)
    return 1.0 - (ratio ** 2) - ((1 - ratio) ** 2)


@jit(nopython=True, nogil=NOGIL)
def sum_absolute_errors(y):
    if len(y) == 0:
        return 0.0

    return np.sum(np.abs(y - np.median(y)))


@jit(nopython=True, nogil=NOGIL)
def adversarial_sum_absolute_errors(y_l, y_r, y_i):
    if len(y_i) == 0:
        return sum_absolute_errors(y_l) + sum_absolute_errors(y_r), (0, 0)

    y_i = np.sort(y_i)
    max_error = 0
    indices = None
    for i in range(len(y_i)):
        error = 0
        error += sum_absolute_errors(np.concatenate((y_l, y_i[:i])))
        error += sum_absolute_errors(np.concatenate((y_r, y_i[i:])))

        if error > max_error:
            max_error = error
            indices = (0, i)

    for i in range(len(y_i)):
        error = 0
        error += sum_absolute_errors(np.concatenate((y_l, y_i[i:])))
        error += sum_absolute_errors(np.concatenate((y_r, y_i[:i])))

        if error > max_error:
            max_error = error
            indices = (i, len(y_i))

    return max_error, indices


@jit(nopython=True, nogil=NOGIL)
def chen_adversarial_sum_absolute_errors(y_l, y_li, y_ri, y_r):
    if len(y_li) == 0 and len(y_ri) == 0:
        return sum_absolute_errors(y_l) + sum_absolute_errors(y_r), 1

    s1 = sum_absolute_errors(np.concatenate((y_l, y_li))) + sum_absolute_errors(
        np.concatenate((y_r, y_ri))
    )
    s2 = sum_absolute_errors(y_l) + sum_absolute_errors(
        np.concatenate((y_li, y_ri, y_r))
    )
    s3 = sum_absolute_errors(np.concatenate((y_l, y_li, y_ri))) + sum_absolute_errors(
        y_r
    )
    s4 = sum_absolute_errors(np.concatenate((y_l, y_ri))) + sum_absolute_errors(
        np.concatenate((y_r, y_li))
    )

    worst_case = max(s1, s2, s3, s4)

    if s1 == worst_case:
        return s1, 1
    elif s2 == worst_case:
        return s2, 2
    elif s3 == worst_case:
        return s3, 3
    else:
        return s4, 4


@jit(nopython=True, nogil=NOGIL)
def weighted_gini(l_0, l_1, r_0, r_1):
    l_t = l_0 + l_1
    r_t = r_0 + r_1

    # Prevent division by 0
    if l_t == 0:
        l_p = 1.0
    else:
        l_p = l_0 / (l_0 + l_1)
    if r_t == 0:
        r_p = 1.0
    else:
        r_p = r_0 / (r_0 + r_1)

    gini = l_t * (1 - (l_p ** 2) - ((1 - l_p) ** 2)) + r_t * (
        1 - (r_p ** 2) - ((1 - r_p) ** 2)
    )

    total = l_t + r_t
    if total != 0:
        gini /= total
        return gini
    else:
        return 1.0


@jit(nopython=True, nogil=NOGIL)
def _counts_to_one_class_adv_gini(counts, rho, chen_heuristic):
    # Apply rho by moving a number of samples back from intersect
    rho_inv = 1.0 - rho
    left_mal = counts[LEFT][1] + int(round(rho_inv * counts[LEFT_INTERSECT][1]))
    right_mal = counts[RIGHT][1] + int(round(rho_inv * counts[RIGHT_INTERSECT][1]))
    left_i_mal = int(round(rho * counts[LEFT_INTERSECT][1]))
    right_i_mal = int(round(rho * counts[RIGHT_INTERSECT][1]))

    # Compute the adversarial gini gain
    if chen_heuristic:
        adv_gini, _ = chen_adversarial_gini_gain_one_class(
            counts[LEFT][0],
            left_mal,
            counts[RIGHT][0],
            right_mal,
            left_i_mal,
            right_i_mal,
        )
    else:
        adv_gini, _ = adversarial_gini_gain_one_class(
            counts[LEFT][0],
            left_mal,
            counts[RIGHT][0],
            right_mal,
            left_i_mal + right_i_mal,
        )

    return adv_gini


@jit(nopython=True, nogil=NOGIL)
def _counts_to_two_class_adv_gini(counts, rho, chen_heuristic):
    # Apply rho by moving a number of samples back from intersect
    rho_inv = 1.0 - rho
    left = counts[LEFT] + np.rint(rho_inv * counts[LEFT_INTERSECT]).astype(np.int64)
    right = counts[RIGHT] + np.rint(rho_inv * counts[RIGHT_INTERSECT]).astype(np.int64)
    left_i = np.rint(rho * counts[LEFT_INTERSECT]).astype(np.int64)
    right_i = np.rint(rho * counts[RIGHT_INTERSECT]).astype(np.int64)

    # Compute the adversarial gini gain
    if chen_heuristic:
        adv_gini, _, _ = chen_adversarial_gini_gain_two_class(
            left[0],
            left[1],
            left_i[0],
            left_i[1],
            right_i[0],
            right_i[1],
            right[0],
            right[1],
        )
    else:
        adv_gini, _, _ = adversarial_gini_gain_two_class(
            left[0],
            left[1],
            left_i[0],
            left_i[1],
            right_i[0],
            right_i[1],
            right[0],
            right[1],
        )

    return adv_gini


class BaseGrootTree(BaseEstimator):
    """
    Base class for GROOT decision trees.

    Implements high level fitting operation and exporting to strings/JSON.
    """

    def __init__(
        self,
        max_depth=5,
        min_samples_split=2,
        min_samples_leaf=1,
        max_features=None,
        robust_weight=1.0,
        attack_model=None,
        chen_heuristic=False,
        random_state=None,
    ):
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.robust_weight = robust_weight
        self.attack_model = attack_model
        self.chen_heuristic = chen_heuristic
        self.random_state = random_state

    def fit(self, X, y, check_input=True):
        """
        Build a robust and fair binary decision tree from the training set
        (X, y) using greedy splitting according to the weighted adversarial
        Gini impurity and fairness impurity.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The training samples.
        y : array-like of shape (n_samples,)
            The class labels as integers 0 (benign) or 1 (malicious)

        Returns
        -------
        self : object
            Fitted estimator.
        """

        if check_input:
            X, y = check_X_y(X, y)
            y = self._check_target(y)

        self.n_samples_, self.n_features_in_ = X.shape

        if self.attack_model is None:
            attack_model = [""] * X.shape[1]
        else:
            attack_model = self.attack_model

        # Turn numerical features in attack model into tuples to make fitting
        # code simpler
        self.attack_model_ = _attack_model_to_tuples(attack_model)

        self.random_state_ = check_random_state(self.random_state)

        if self.max_features == "sqrt":
            self.max_features_ = int(np.sqrt(self.n_features_in_))
        elif self.max_features == "log2":
            self.max_features_ = int(np.log2(self.n_features_in_))
        elif self.max_features is None:
            self.max_features_ = self.n_features_in_
        else:
            self.max_features_ = self.max_features

        if self.max_features_ == 0:
            self.max_features_ = 1

        # Keep track of the minimum and maximum split value for each feature
        constraints = np.concatenate(
            (np.min(X, axis=0).reshape(-1, 1), np.max(X, axis=0).reshape(-1, 1)), axis=1
        )

        self.root_ = self.__fit_recursive(X, y, constraints)

        return self

    def __fit_recursive(self, X, y, constraints, depth=0):
        """
        Recursively fit the decision tree on the training dataset (X, y).

        The constraints make sure that leaves are well formed, e.g. don't
        cross an earlier split. Stop when the depth has reached self.max_depth,
        when a leaf is pure or when the leaf contains too few samples.
        """
        if (
            (self.max_depth is not None and depth == self.max_depth)
            or len(y) < self.min_samples_split
            or np.all(y == y[0])
        ):
            return self._create_leaf(y)

        current_score = self._score(y)

        rule, feature, split_score = self.__best_adversarial_decision(X, y, constraints)

        score_gain = current_score - split_score

        if rule is None or score_gain <= 0.00:
            return self._create_leaf(y)

        # Assert that the split obeys constraints made by previous splits
        assert rule >= constraints[feature][0]
        assert rule < constraints[feature][1]

        X_left, y_left, X_right, y_right = self._split_left_right(
            X,
            y,
            rule,
            feature,
        )

        if len(y_left) < self.min_samples_leaf or len(y_right) < self.min_samples_leaf:
            return self._create_leaf(y)

        # Set the right bound and store old one for after recursion
        old_right_bound = constraints[feature][1]
        constraints[feature][1] = rule

        left_node = self.__fit_recursive(X_left, y_left, constraints, depth + 1)

        # Reset right bound, set left bound, store old one for after recursion
        constraints[feature][1] = old_right_bound
        old_left_bound = constraints[feature][0]
        constraints[feature][0] = rule

        right_node = self.__fit_recursive(X_right, y_right, constraints, depth + 1)

        # Reset the left bound
        constraints[feature][0] = old_left_bound

        node = NumericalNode(feature, rule, left_node, right_node, _TREE_UNDEFINED)

        return node

    def __best_adversarial_decision(self, X, y, constraints):
        """
        Find the best split by iterating through each feature and scanning
        it for that feature's optimal split.
        """

        best_score = 10e9
        best_rule = None
        best_feature = None

        # If there is a limit on features to consider in a split then choose
        # that number of random features.
        all_features = np.arange(self.n_features_in_)
        features = self.random_state_.choice(
            all_features, size=self.max_features_, replace=False
        )

        for feature in features:
            score, decision_rule = self._scan_feature(X, y, feature, constraints)

            if decision_rule is not None and score < best_score:
                best_score = score
                best_rule = decision_rule
                best_feature = feature

        return best_rule, best_feature, best_score

    def to_string(self):
        result = ""
        result += f"Parameters: {self.get_params()}\n"

        if hasattr(self, "root_"):
            result += f"Tree:\n{self.root_.pretty_print()}"
        else:
            result += "Tree has not yet been fitted"

        return result

    def to_json(self, output_file="tree.json"):
        dictionary = {
            "params": self.get_params(),
        }
        if hasattr(self, "root_"):
            dictionary["tree"] = self.root_.to_json()
        else:
            dictionary["tree"] = None

        if output_file is None:
            return dictionary
        else:
            with open(output_file, "w") as fp:
                json.dump(dictionary, fp, indent=2, default=convert_numpy)

    def to_xgboost_json(self, output_file="tree.json"):
        check_is_fitted(self, "root_")

        dictionary, _ = self.root_.to_xgboost_json(0, 0)

        if output_file is None:
            return dictionary
        else:
            with open(output_file, "w") as fp:
                # If saving to file then surround dict in list brackets
                json.dump([dictionary], fp, indent=2, default=convert_numpy)


class GrootTreeClassifier(BaseGrootTree, ClassifierMixin):
    """
    A robust decision tree for binary classification.
    """

    def __init__(
        self,
        max_depth=5,
        min_samples_split=2,
        min_samples_leaf=1,
        max_features=None,
        robust_weight=1.0,
        attack_model=None,
        one_adversarial_class=False,
        chen_heuristic=False,
        random_state=None,
    ):
        """
        Parameters
        ----------
        max_depth : int, optional
            The maximum depth for the decision tree once fitted.
        min_samples_split : int, optional
            The minimum number of samples required to split a node.
        min_samples_leaf : int, optional
            The minimum number of samples required to make a leaf.
        max_features : int or {"sqrt", "log2"}, optional
            The number of features to consider while making each split, if None then all features are considered.
        robust_weight : float, optional
            The ratio of samples that are actually moved by an adversary.
        attack_model : array-like of shape (n_features,), optional
            Attacker capabilities for perturbing X. By default, all features are considered not perturbable.
        one_adversarial_class : bool, optional
            Whether one class (malicious, 1) perturbs their samples or if both classes (benign and malicious, 0 and 1) do so.
        chen_heuristic : bool, optional
            Whether to use the heuristic for the adversarial Gini impurity from Chen et al. (2019) instead of GROOT's adversarial Gini impurity.
        random_state : int, optional
            Controls the sampling of the features to consider when looking for the best split at each node.

        Attributes
        ----------
        classes_ : ndarray of shape (n_classes,)
            The class labels.
        max_features_ : int
            The inferred value of max_features.
        n_samples_ : int
            The number of samples when `fit` is performed.
        n_features_ : int
            The number of features when `fit` is performed.
        root_ : Node
            The root node of the tree after fitting.
        """
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.robust_weight = robust_weight
        self.attack_model = attack_model
        self.one_adversarial_class = one_adversarial_class
        self.chen_heuristic = chen_heuristic
        self.random_state = random_state

    def _check_target(self, y):
        target_type = type_of_target(y)
        if target_type != "binary":
            raise ValueError(
                f"Unknown label type: classifier only supports binary labels but found {target_type}"
            )

        self.classes_, y = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)

        return y

    def _score(self, y):
        return gini_impurity(np.sum(y == 0), np.sum(y == 1))

    def _create_leaf(self, y):
        """
        Create a leaf object that predicts according to the ratio of benign
        and malicious labels in the array y.
        """

        # Count the number of points that fall into this leaf including
        # adversarially moved points
        label_counts = np.bincount(y, minlength=2)

        # Set the leaf's prediction value to the weighted average of the
        # prediction with and without moving points
        value = label_counts / np.sum(label_counts)

        return Node(_TREE_UNDEFINED, _TREE_LEAF, _TREE_LEAF, value)

    def _scan_feature(self, X, y, feature, constraints):
        """
        Scan feature to find the locally optimal split.
        """

        samples = X[:, feature]
        attack_mode = self.attack_model_[feature]
        constraint = constraints[feature]

        # If possible, use the faster scan implementation
        if self.robust_weight == 1:
            return _scan_numerical_feature_fast(
                samples,
                y,
                *attack_mode,
                *constraint,
                self.chen_heuristic,
                self.one_adversarial_class,
            )
        else:
            return self.__scan_feature_numerical(samples, y, attack_mode, *constraint)

    def __initialize_scan(self, samples, y, attack_mode):
        queue = []
        counts = np.array(
            [[0, 0], [0, 0], [0, 0], [0, 0]],
            dtype=np.int64,
        )

        if attack_mode == "":
            counts[RIGHT] = np.bincount(y)

            for sample, label in zip(samples, y):
                queue.append((sample, label, RIGHT, LEFT))

        elif attack_mode == ">":
            counts[RIGHT] = np.bincount(y)

            for sample, label in zip(samples, y):
                if label == 0 and self.one_adversarial_class:
                    queue.append((sample, label, RIGHT, LEFT))
                else:
                    queue.append((sample, label, RIGHT, LEFT_INTERSECT))

        elif attack_mode == "<":
            if self.one_adversarial_class:
                counts[RIGHT][0] = np.sum(y == 0)
                counts[RIGHT_INTERSECT][1] = np.sum(y == 1)
            else:
                counts[RIGHT_INTERSECT] = np.bincount(y)

            for sample, label in zip(samples, y):
                if label == 0 and self.one_adversarial_class:
                    queue.append((sample, label, RIGHT, LEFT))
                else:
                    queue.append((sample, label, RIGHT_INTERSECT, LEFT))

        elif attack_mode == "<>":
            if self.one_adversarial_class:
                counts[RIGHT][0] = np.sum(y == 0)
                counts[RIGHT_INTERSECT][1] = np.sum(y == 1)
            else:
                counts[RIGHT_INTERSECT] = np.bincount(y)

            for sample, label in zip(samples, y):
                if label == 0 and self.one_adversarial_class:
                    queue.append((sample, label, RIGHT, LEFT))
                else:
                    queue.append((sample, label, RIGHT_INTERSECT, LEFT_INTERSECT))

        elif isinstance(attack_mode, numbers.Number):
            counts[RIGHT] = np.bincount(y)

            for sample, label in zip(samples, y):
                if label == 0 and self.one_adversarial_class:
                    queue.append((sample, label, RIGHT, LEFT))
                else:
                    queue.append(
                        (sample - attack_mode, label, RIGHT, RIGHT_INTERSECT),
                    )
                    queue.append(
                        (sample, label, RIGHT_INTERSECT, LEFT_INTERSECT),
                    )
                    queue.append(
                        (sample + attack_mode, label, LEFT_INTERSECT, LEFT),
                    )

        elif isinstance(attack_mode, tuple):
            counts[RIGHT] = np.bincount(y)

            for sample, label in zip(samples, y):
                if label == 0 and self.one_adversarial_class:
                    queue.append((sample, label, RIGHT, LEFT))
                else:
                    queue.append(
                        (sample - attack_mode[0], label, RIGHT, RIGHT_INTERSECT),
                    )
                    queue.append(
                        (sample, label, RIGHT_INTERSECT, LEFT_INTERSECT),
                    )
                    queue.append(
                        (sample + attack_mode[1], label, LEFT_INTERSECT, LEFT),
                    )

        # Sort queue in reverse order since popping from end is faster
        queue.sort(reverse=True)

        return queue, counts

    def __scan_feature_numerical(
        self,
        samples,
        y,
        attack_mode,
        left_bound,
        right_bound,
    ):
        """
        Scan a numerical feature for the optimal split by identifying every
        potential split, sorting these and iterating through them.

        While iterating from the left to the right, remember counts for the
        right / right_intersect / left_intersect / left positions and label.
        """
        best_score = 10e9
        best_split = None

        queue, counts = self.__initialize_scan(samples, y, attack_mode)

        while len(queue) > 0 and queue[-1][0] < left_bound:
            point, label, move_from, move_to = queue.pop()
            counts[move_from][label] -= 1
            counts[move_to][label] += 1

        adv_gini = None
        while queue:
            point, label, move_from, move_to = queue.pop()
            counts[move_from][label] -= 1
            counts[move_to][label] += 1

            if point >= right_bound:
                break

            if len(queue) > 0 and queue[-1][0] != point:
                # Compute the adversarial Gini gain
                if self.one_adversarial_class:
                    adv_gini = _counts_to_one_class_adv_gini(
                        counts, self.robust_weight, self.chen_heuristic
                    )
                else:
                    adv_gini = _counts_to_two_class_adv_gini(
                        counts, self.robust_weight, self.chen_heuristic
                    )

                # Maximize the margin of the split
                split = (point + queue[-1][0]) * 0.5

                if (
                    adv_gini is not None
                    and adv_gini < best_score
                    and split < right_bound
                ):
                    best_score = adv_gini
                    best_split = split

            if len(queue) == 0:
                break

        return best_score, best_split

    def _split_left_right(self, X, y, rule, feature):
        """
        Split the dataset (X, y) into a left and right dataset according to the
        optimal split determined by rule and feature.

        Only a ratio 'robust_weight' of moving (malicious) points are actually
        transfered to the other side.
        """

        # Get perturbation range for this feature
        dec, inc = self.attack_model_[feature]

        if self.one_adversarial_class:
            # Determine the indices of samples on each side of the split
            label_0 = y == 0
            label_1 = np.invert(label_0)

            i_left = np.where(
                (label_0 & (X[:, feature] <= rule))
                | (label_1 & (X[:, feature] + inc <= rule))
            )[0]
            i_left_intersection = np.where(
                label_1 & (X[:, feature] + inc > rule) & (X[:, feature] <= rule)
            )[0]
            i_right_intersection = np.where(
                label_1 & (X[:, feature] > rule) & (X[:, feature] - dec <= rule)
            )[0]
            i_right = np.where(
                (label_0 & (X[:, feature] > rule))
                | (label_1 & (X[:, feature] - dec > rule))
            )[0]
        else:
            # Determine the indices of samples on each side of the split
            i_left = np.where(X[:, feature] + inc <= rule)[0]
            i_left_intersection = np.where(
                (X[:, feature] + inc > rule) & (X[:, feature] <= rule)
            )[0]
            i_right_intersection = np.where(
                (X[:, feature] > rule) & (X[:, feature] - dec <= rule)
            )[0]
            i_right = np.where(X[:, feature] - dec > rule)[0]

        # Count samples with labels 0 and 1 left and right
        l_0, l_1 = np.bincount(y[i_left], minlength=2)
        r_0, r_1 = np.bincount(y[i_right], minlength=2)

        # Determine labels on the left and right intersection
        y_left_intersection = y[i_left_intersection]
        y_right_intersection = y[i_right_intersection]

        i_left_intersection_0 = np.where(y_left_intersection == 0)[0]
        i_left_intersection_1 = np.where(y_left_intersection == 1)[0]
        i_right_intersection_0 = np.where(y_right_intersection == 0)[0]
        i_right_intersection_1 = np.where(y_right_intersection == 1)[0]

        li_0 = len(i_left_intersection_0)
        li_1 = len(i_left_intersection_1)
        ri_0 = len(i_right_intersection_0)
        ri_1 = len(i_right_intersection_1)

        # Compute optimal movement
        if self.one_adversarial_class:
            # Compute numbers of samples after applying rho
            assert li_0 == 0
            assert ri_0 == 0
            l_1 = l_1 + round((1.0 - self.robust_weight) * li_1)
            r_1 = r_1 + round((1.0 - self.robust_weight) * ri_1)
            li_1 = round(self.robust_weight * li_1)
            ri_1 = round(self.robust_weight * ri_1)
            i_1 = li_1 + ri_1

            # Determine optimal movement
            if self.chen_heuristic:
                _, m1 = chen_adversarial_gini_gain_one_class(
                    l_0, l_1, r_0, r_1, li_1, ri_1
                )
            else:
                _, m1 = adversarial_gini_gain_one_class(l_0, l_1, r_0, r_1, i_1)
            m0 = None
        else:
            # Compute numbers of samples after applying rho
            l_0 = l_0 + round((1.0 - self.robust_weight) * li_0)
            l_1 = l_1 + round((1.0 - self.robust_weight) * li_1)
            r_0 = r_0 + round((1.0 - self.robust_weight) * ri_0)
            r_1 = r_1 + round((1.0 - self.robust_weight) * ri_1)
            li_0 = round(self.robust_weight * li_0)
            li_1 = round(self.robust_weight * li_1)
            ri_0 = round(self.robust_weight * ri_0)
            ri_1 = round(self.robust_weight * ri_1)

            # Determine optimal movement
            if self.chen_heuristic:
                _, m1, m0 = chen_adversarial_gini_gain_two_class(
                    l_0, l_1, li_0, li_1, ri_0, ri_1, r_0, r_1
                )
            else:
                _, m1, m0 = adversarial_gini_gain_two_class(
                    l_0, l_1, li_0, li_1, ri_0, ri_1, r_0, r_1
                )

        # Move label 1 samples according to m1
        if m1 > li_1:
            n_move_left = m1 - li_1

            i_left_intersection_1 = np.concatenate(
                (i_left_intersection_1, i_right_intersection_1[:n_move_left])
            )
            i_right_intersection_1 = i_right_intersection_1[n_move_left:]
        elif m1 < li_1:
            n_move_right = li_1 - m1

            i_left_intersection_1 = i_left_intersection_1[n_move_right:]
            i_right_intersection_1 = np.concatenate(
                (i_left_intersection_1[:n_move_right], i_right_intersection_1)
            )

        # Move label 0 samples according to m0 (not used if one_adversarial_class=True)
        if m0:
            if m0 > li_0:
                n_move_left = m0 - li_0

                i_left_intersection_0 = np.concatenate(
                    (i_left_intersection_0, i_right_intersection_0[:n_move_left])
                )
                i_right_intersection_0 = i_right_intersection_0[n_move_left:]
            elif m0 < li_0:
                n_move_right = li_0 - m0

                i_left_intersection_0 = i_left_intersection_0[n_move_right:]
                i_right_intersection_0 = np.concatenate(
                    (i_left_intersection_0[:n_move_right], i_right_intersection_0)
                )

        i_left = np.concatenate(
            (
                i_left,
                i_left_intersection_0,
                i_left_intersection_1,
            )
        )

        i_right = np.concatenate(
            (
                i_right,
                i_right_intersection_0,
                i_right_intersection_1,
            )
        )

        # Move the resulting intersection arrays to the left and right sides
        X_left = X[i_left]
        y_left = y[i_left]
        X_right = X[i_right]
        y_right = y[i_right]

        return X_left, y_left, X_right, y_right

    def predict_proba(self, X):
        """
        Predict class probabilities of the input samples X.

        The class probability is the fraction of samples of the same class in
        the leaf.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to predict.

        Returns
        -------
        proba : array of shape (n_samples,)
            The probability for each input sample of being malicious.
        """

        check_is_fitted(self, "root_")

        X = check_array(X)
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "Received different number of features during predict than during fit"
            )

        predictions = []
        for sample in X:
            predictions.append(self.root_.predict(sample))

        predictions = np.array(predictions)
        predictions /= np.sum(predictions, axis=1)[:, np.newaxis]

        return predictions

    def predict(self, X):
        """
        Predict the classes of the input samples X.

        The predicted class is the most frequently occuring class label in a
        leaf.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to predict.

        Returns
        -------
        y : array-like of shape (n_samples,)
            The predicted class labels
        """

        y_pred_proba = self.predict_proba(X)

        return self.classes_.take(np.argmax(y_pred_proba, axis=1))


class GrootTreeRegressor(BaseGrootTree, RegressorMixin):
    """
    A robust decision tree for single target regression.

    Optimizes the adversarial sum of absolute errors at each split.
    """

    def __init__(
        self,
        max_depth=5,
        min_samples_split=2,
        min_samples_leaf=1,
        max_features=None,
        robust_weight=1.0,
        attack_model=None,
        chen_heuristic=False,
        random_state=None,
    ):
        """
        Parameters
        ----------
        max_depth : int, optional
            The maximum depth for the decision tree once fitted.
        min_samples_split : int, optional
            The minimum number of samples required to split a node.
        min_samples_leaf : int, optional
            The minimum number of samples required to make a leaf.
        max_features : int or {"sqrt", "log2"}, optional
            The number of features to consider while making each split, if None then all features are considered.
        robust_weight : float, optional
            The ratio of samples that are actually moved by an adversary.
        attack_model : array-like of shape (n_features,), optional
            Attacker capabilities for perturbing X. By default, all features are considered not perturbable.
        one_adversarial_class : bool, optional
            Whether one class (malicious, 1) perturbs their samples or if both classes (benign and malicious, 0 and 1) do so.
        chen_heuristic : bool, optional
            Whether to use the heuristic for the adversarial Gini impurity from Chen et al. (2019) instead of GROOT's adversarial Gini impurity.
        random_state : int, optional
            Controls the sampling of the features to consider when looking for the best split at each node.

        Attributes
        ----------
        classes_ : ndarray of shape (n_classes,)
            The class labels.
        max_features_ : int
            The inferred value of max_features.
        n_samples_ : int
            The number of samples when `fit` is performed.
        n_features_ : int
            The number of features when `fit` is performed.
        root_ : Node
            The root node of the tree after fitting.
        """
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.robust_weight = robust_weight
        self.attack_model = attack_model
        self.chen_heuristic = chen_heuristic
        self.random_state = random_state

    def _check_target(self, y):
        target_type = type_of_target(y)
        if target_type not in {"continuous", "multiclass", "binary"}:
            raise ValueError(
                f"Unknown label type: regressor only supports continuous/multiclass/binary targets but found {target_type}"
            )

        # Make a copy of y if it is readonly to prevent errors
        if not y.flags.writeable:
            y = np.copy(y)

        return y

    def _score(self, y):
        return sum_absolute_errors(y)

    def _create_leaf(self, y):
        """
        Create a leaf object that predicts according to the ratio of benign
        and malicious labels in the array y.
        """

        # For optimal mean absolute error predict the median
        value = np.median(y)

        return Node(_TREE_UNDEFINED, _TREE_LEAF, _TREE_LEAF, value)

    def _scan_feature(self, X, y, feature, constraints):
        """
        Scan a feature to find the locally optimal split.
        """

        samples = X[:, feature]
        attack_mode = self.attack_model_[feature]
        constraint = constraints[feature]

        # If possible, use the faster scan implementation
        if self.robust_weight == 1:
            return _scan_numerical_feature_fast_regression(
                samples, y, *attack_mode, *constraint, self.chen_heuristic
            )
        else:
            raise ValueError("Robust weight not yet supported")

    def _split_left_right(self, X, y, rule, feature):
        """
        Split the dataset (X, y) into a left and right dataset according to the
        optimal split determined by rule and feature.

        Only a ratio 'robust_weight' of moving (malicious) points are actually
        transfered to the other side.
        """

        # Get perturbation range for this feature
        dec, inc = self.attack_model_[feature]

        # Determine the indices of samples on each side of the split
        i_left = np.where(X[:, feature] + inc <= rule)[0]
        i_right = np.where(X[:, feature] - dec > rule)[0]

        y_left = y[i_left]
        y_right = y[i_right]

        if self.chen_heuristic:
            i_left_intersection = np.where(
                (X[:, feature] + inc > rule) & (X[:, feature] <= rule)
            )[0]
            i_right_intersection = np.where(
                (X[:, feature] - dec <= rule) & (X[:, feature] > rule)
            )[0]

            y_left_intersection = y[i_left_intersection]
            y_right_intersection = y[i_right_intersection]

            # Compute optimal movement
            _, configuration = chen_adversarial_sum_absolute_errors(
                y_left,
                y_left_intersection,
                y_right_intersection,
                y_right,
            )

            # Move the resulting intersection arrays to the left and right sides
            if configuration == 1:
                i_left = np.concatenate((i_left, i_left_intersection))
                i_right = np.concatenate((i_right, i_right_intersection))
            elif configuration == 2:
                i_right = np.concatenate(
                    (i_right, i_left_intersection, i_right_intersection)
                )
            elif configuration == 3:
                i_left = np.concatenate(
                    (i_left, i_left_intersection, i_right_intersection)
                )
            else:
                i_left = np.concatenate((i_left, i_right_intersection))
                i_right = np.concatenate((i_right, i_left_intersection))
        else:
            i_intersection = np.where(
                (X[:, feature] + inc > rule) & (X[:, feature] - dec <= rule)
            )[0]
            y_intersection = np.sort(y[i_intersection])

            # Compute optimal movement
            _, (l_start, l_end) = adversarial_sum_absolute_errors(
                y_left, y_right, y_intersection
            )
            if l_start == 0:
                r_start = l_end
                r_end = -1
            else:
                r_start = 0
                r_end = l_start

            # Move the resulting intersection arrays to the left and right sides
            i_left = np.concatenate((i_left, i_intersection[l_start:l_end]))
            i_right = np.concatenate((i_right, i_intersection[r_start:r_end]))

        X_left = X[i_left]
        y_left = y[i_left]
        X_right = X[i_right]
        y_right = y[i_right]

        return X_left, y_left, X_right, y_right

    def predict(self, X):
        """
        Predict the targets of the input samples X.

        The predicted value is the median of values in a leaf.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to predict.

        Returns
        -------
        y : array-like of shape (n_samples,)
            The predicted class labels
        """

        check_is_fitted(self, "root_")

        X = check_array(X)
        if X.shape[1] != self.n_features_in_:
            raise ValueError(
                "Received different number of features during predict than during fit"
            )

        predictions = []
        for sample in X:
            predictions.append(self.root_.predict(sample))

        return np.array(predictions)


def _build_tree_parallel(base_tree, X, y, indices, seed, n_samples):
    tree = clone(base_tree)

    random_state = check_random_state(seed)
    tree.random_state = random_state

    i_bootstrap = random_state.choice(indices, n_samples)
    X_bootstrap = X[i_bootstrap]
    y_bootstrap = y[i_bootstrap]

    tree.fit(X_bootstrap, y_bootstrap, check_input=False)
    return tree


class BaseGrootRandomForest(BaseEstimator):
    """
    Base class for robust (GROOT) random forests.
    """

    def __init__(
        self,
        n_estimators=100,
        max_depth=None,
        max_features="sqrt",
        min_samples_split=2,
        min_samples_leaf=1,
        robust_weight=1.0,
        attack_model=None,
        verbose=False,
        chen_heuristic=False,
        max_samples=None,
        n_jobs=None,
        random_state=None,
    ):
        """
        Parameters
        ----------
        n_estimators : int, optional
            The number of decision trees to fit in the forest.
        max_depth : int, optional
            The maximum depth for the decision trees once fitted.
        max_features : int or {"sqrt", "log2", None}, optional
            The number of features to consider while making each split, if None then all features are considered.
        min_samples_split : int, optional
            The minimum number of samples required to split a tree node.
        min_samples_leaf : int, optional
            The minimum number of samples required to make a tree leaf.
        robust_weight : float, optional
            The ratio of samples that are actually moved by an adversary.
        attack_model : array-like of shape (n_features,), optional
            Attacker capabilities for perturbing X. The attack model needs to describe for every feature in which way it can be perturbed.
        verbose : bool, optional
            Whether to print fitting progress on screen.
        chen_heuristic : bool, optional
            Whether to use the heuristic for the adversarial Gini impurity from Chen et al. (2019) instead of GROOT's adversarial Gini impurity.
        max_samples : float, optional
            The fraction of samples to draw from X to train each decision tree. If None (default), then draw X.shape[0] samples.
        n_jobs : int, optional
            The number of jobs to run in parallel when fitting trees. See joblib.
        random_state : int, optional
            Controls the sampling of the features to consider when looking for the best split at each node.

        Attributes
        ----------
        estimators_ : list of GrootTree
            The collection of fitted sub-estimators.
        n_samples_ : int
            The number of samples when `fit` is performed.
        n_features_ : int
            The number of features when `fit` is performed.
        """
        self.n_estimators = n_estimators
        self.max_depth = max_depth
        self.max_features = max_features
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.robust_weight = robust_weight
        self.attack_model = attack_model
        self.verbose = verbose
        self.chen_heuristic = chen_heuristic
        self.max_samples = max_samples
        self.n_jobs = n_jobs
        self.random_state = random_state

    def fit(self, X, y):
        """
        Build a robust and fair random forest of binary decision trees from
        the training set (X, y) using greedy splitting according to the
        adversarial Gini combined with fair gini under the given attack model.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The training samples.
        y : array-like of shape (n_samples,)
            The class labels as integers 0 (benign) or 1 (malicious)

        Returns
        -------
        self : object
            Fitted estimator.
        """

        X, y = check_X_y(X, y)
        y = self._check_target(y)

        self.n_samples_, self.n_features_in_ = X.shape

        random_state = check_random_state(self.random_state)

        # Generate seeds for the random states of each tree to prevent each
        # of them from fitting exactly the same way, but use the random state
        # to keep the forest reproducible
        seeds = [
            random_state.randint(np.iinfo(np.int32).max)
            for _ in range(self.n_estimators)
        ]

        tree = self._get_estimator()

        if self.max_samples:
            n_bootstrap_samples = int(self.n_samples_ * self.max_samples)
        else:
            n_bootstrap_samples = self.n_samples_

        indices = np.arange(X.shape[0])
        self.estimators_ = Parallel(
            n_jobs=self.n_jobs, verbose=self.verbose, prefer="processes"
        )(
            delayed(_build_tree_parallel)(
                tree, X, y, indices, seed, n_bootstrap_samples
            )
            for seed in seeds
        )

        return self

    def __str__(self):
        result = ""
        result += f"Parameters: {self.get_params()}\n"

        if hasattr(self, "estimators_"):
            for tree in self.estimators_:
                result += f"Tree:\n{tree.root_.pretty_print()}\n"
        else:
            result += "Forest has not yet been fitted"

        return result

    def to_json(self, output_file="forest.json"):
        with open(output_file, "w") as fp:
            dictionary = {
                "params": self.get_params(),
            }
            if hasattr(self, "estimators_"):
                dictionary["trees"] = [tree.to_json(None) for tree in self.estimators_]
                json.dump(dictionary, fp, indent=2, default=convert_numpy)
            else:
                dictionary["trees"] = None
                json.dump(dictionary, fp)

    def to_xgboost_json(self, output_file="forest.json"):
        if hasattr(self, "estimators_"):
            dictionary = [tree.to_xgboost_json(None) for tree in self.estimators_]

            if output_file:
                with open(output_file, "w") as fp:
                    json.dump(dictionary, fp, indent=2, default=convert_numpy)
            else:
                return dictionary
        else:
            raise Exception("Forest not yet fitted")


class GrootRandomForestClassifier(BaseGrootRandomForest, ClassifierMixin):
    """
    A robust random forest for binary classification.
    """

    def __init__(
        self,
        n_estimators=100,
        max_depth=None,
        max_features="sqrt",
        min_samples_split=2,
        min_samples_leaf=1,
        robust_weight=1.0,
        attack_model=None,
        one_adversarial_class=False,
        verbose=False,
        chen_heuristic=False,
        max_samples=None,
        n_jobs=None,
        random_state=None,
    ):
        """
        Parameters
        ----------
        n_estimators : int, optional
            The number of decision trees to fit in the forest.
        max_depth : int, optional
            The maximum depth for the decision trees once fitted.
        max_features : int or {"sqrt", "log2", None}, optional
            The number of features to consider while making each split, if None then all features are considered.
        min_samples_split : int, optional
            The minimum number of samples required to split a tree node.
        min_samples_leaf : int, optional
            The minimum number of samples required to make a tree leaf.
        robust_weight : float, optional
            The ratio of samples that are actually moved by an adversary.
        attack_model : array-like of shape (n_features,), optional
            Attacker capabilities for perturbing X. The attack model needs to describe for every feature in which way it can be perturbed.
        one_adversarial_class : bool, optional
            Whether one class (malicious, 1) perturbs their samples or if both classes (benign and malicious, 0 and 1) do so.
        verbose : bool, optional
            Whether to print fitting progress on screen.
        chen_heuristic : bool, optional
            Whether to use the heuristic for the adversarial Gini impurity from Chen et al. (2019) instead of GROOT's adversarial Gini impurity.
        max_samples : float, optional
            The fraction of samples to draw from X to train each decision tree. If None (default), then draw X.shape[0] samples.
        n_jobs : int, optional
            The number of jobs to run in parallel when fitting trees. See joblib.
        random_state : int, optional
            Controls the sampling of the features to consider when looking for the best split at each node.

        Attributes
        ----------
        estimators_ : list of GrootTree
            The collection of fitted sub-estimators.
        n_samples_ : int
            The number of samples when `fit` is performed.
        n_features_ : int
            The number of features when `fit` is performed.
        """
        super().__init__(
            n_estimators=n_estimators,
            max_depth=max_depth,
            max_features=max_features,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            robust_weight=robust_weight,
            attack_model=attack_model,
            verbose=verbose,
            chen_heuristic=chen_heuristic,
            max_samples=max_samples,
            n_jobs=n_jobs,
            random_state=random_state,
        )
        self.one_adversarial_class = one_adversarial_class

    def _check_target(self, y):
        target_type = type_of_target(y)
        if target_type != "binary":
            raise ValueError(
                f"Unknown label type: classifier only supports binary labels but found {target_type}"
            )

        self.classes_, y = np.unique(y, return_inverse=True)
        self.n_classes_ = len(self.classes_)

        return y

    def _get_estimator(self):
        return GrootTreeClassifier(
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            max_features=self.max_features,
            robust_weight=self.robust_weight,
            attack_model=self.attack_model,
            one_adversarial_class=self.one_adversarial_class,
            chen_heuristic=self.chen_heuristic,
        )

    def predict_proba(self, X):
        """
        Predict class probabilities of the input samples X.
        The class probability is the average of the probabilities predicted by
        each decision tree. The probability prediction of each tree is the
        fraction of samples of the same class in the leaf.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to predict.

        Returns
        -------
        proba : array of shape (n_samples,)
            The probability for each input sample of being malicious.
        """

        check_is_fitted(self, "estimators_")
        X = check_array(X)

        probability_sum = np.zeros((len(X), 2))

        for tree in self.estimators_:
            probabilities = tree.predict_proba(X)
            probability_sum += probabilities

        return probability_sum / self.n_estimators

    def predict(self, X):
        """
        Predict the classes of the input samples X.
        The predicted class is the rounded average of the class labels in
        each predicted leaf.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to predict.

        Returns
        -------
        y : array-like of shape (n_samples,)
            The predicted class labels
        """

        y_pred_proba = self.predict_proba(X)

        return self.classes_.take(np.argmax(y_pred_proba, axis=1))


class GrootRandomForestRegressor(BaseGrootRandomForest, RegressorMixin):
    """
    A robust random forest for binary classification.
    """

    def __init__(
        self,
        n_estimators=100,
        max_depth=None,
        max_features="sqrt",
        min_samples_split=2,
        min_samples_leaf=1,
        robust_weight=1.0,
        attack_model=None,
        verbose=False,
        chen_heuristic=False,
        max_samples=None,
        n_jobs=None,
        random_state=None,
    ):
        """
        Parameters
        ----------
        n_estimators : int, optional
            The number of decision trees to fit in the forest.
        max_depth : int, optional
            The maximum depth for the decision trees once fitted.
        max_features : int or {"sqrt", "log2", None}, optional
            The number of features to consider while making each split, if None then all features are considered.
        min_samples_split : int, optional
            The minimum number of samples required to split a tree node.
        min_samples_leaf : int, optional
            The minimum number of samples required to make a tree leaf.
        robust_weight : float, optional
            The ratio of samples that are actually moved by an adversary.
        attack_model : array-like of shape (n_features,), optional
            Attacker capabilities for perturbing X. The attack model needs to describe for every feature in which way it can be perturbed.
        verbose : bool, optional
            Whether to print fitting progress on screen.
        chen_heuristic : bool, optional
            Whether to use the heuristic for the adversarial Gini impurity from Chen et al. (2019) instead of GROOT's adversarial Gini impurity.
        max_samples : float, optional
            The fraction of samples to draw from X to train each decision tree. If None (default), then draw X.shape[0] samples.
        n_jobs : int, optional
            The number of jobs to run in parallel when fitting trees. See joblib.
        random_state : int, optional
            Controls the sampling of the features to consider when looking for the best split at each node.

        Attributes
        ----------
        estimators_ : list of GrootTree
            The collection of fitted sub-estimators.
        n_samples_ : int
            The number of samples when `fit` is performed.
        n_features_ : int
            The number of features when `fit` is performed.
        """
        super().__init__(
            n_estimators=n_estimators,
            max_depth=max_depth,
            max_features=max_features,
            min_samples_split=min_samples_split,
            min_samples_leaf=min_samples_leaf,
            robust_weight=robust_weight,
            attack_model=attack_model,
            verbose=verbose,
            chen_heuristic=chen_heuristic,
            max_samples=max_samples,
            n_jobs=n_jobs,
            random_state=random_state,
        )

    def _check_target(self, y):
        target_type = type_of_target(y)
        if target_type not in {"continuous", "multiclass", "binary"}:
            raise ValueError(
                f"Unknown label type: regressor only supports continuous/multiclass/binary targets but found {target_type}"
            )

        # Make a copy of y if it is readonly to prevent errors
        if not y.flags.writeable:
            y = np.copy(y)

        return y

    def _get_estimator(self):
        return GrootTreeRegressor(
            max_depth=self.max_depth,
            min_samples_split=self.min_samples_split,
            min_samples_leaf=self.min_samples_leaf,
            max_features=self.max_features,
            robust_weight=self.robust_weight,
            attack_model=self.attack_model,
            chen_heuristic=self.chen_heuristic,
        )

    def predict(self, X):
        """
        Predict the values of the input samples X.

        The predicted values are the means of all individual predictions.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples to predict.

        Returns
        -------
        y : array-like of shape (n_samples,)
            The predicted class labels
        """

        check_is_fitted(self, "estimators_")
        X = check_array(X)

        predictions_sum = np.zeros(len(X))

        for tree in self.estimators_:
            predictions_sum += tree.predict(X)

        return predictions_sum / self.n_estimators



import pprint
import time

try:
    from gurobipy import *
except:
    raise ImportError(
        """The python GUROBI package could not be imported. 
    To install it, run: python -m pip install -i https://pypi.gurobi.com gurobipy"""
    )

from tqdm import tqdm

import numpy as np

from ..attack import AttackWrapper


"""
This code is for the most part written by Hongge Chen and is taken and
adapted from the repository https://github.com/chenhongge/RobustTrees.

It is an implementation of the MILP attack from:
Kantchelian, Alex, J. Doug Tygar, and Anthony Joseph. "Evasion and hardening of
tree ensemble classifiers." International Conference on Machine Learning.
PMLR, 2016.

Feasibility idea from:
Andriushchenko, Maksym, and Matthias Hein. "Provably robust boosted decision 
stumps and trees against adversarial attacks." arXiv preprint arXiv:1906.03526
(2019).

The changes made were related to:
- Default guard_val, round_digits values
- Removing dependency on xgboost
- Taking only a JSON file as input
- Solving a feasibility encoding for fixed epsilon
- Removing print statements
- Removing excessive model updates
- Only keeping binary classification attacks
"""


GUARD_VAL = 5e-6
ROUND_DIGITS = 6


class node_wrapper(object):
    def __init__(
        self,
        treeid,
        nodeid,
        attribute,
        threshold,
        left_leaves,
        right_leaves,
        root=False,
    ):
        # left_leaves and right_leaves are the lists of leaf indices in self.leaf_v_list
        self.attribute = attribute
        self.threshold = threshold
        self.node_pos = []
        self.leaves_lists = []
        self.add_leaves(treeid, nodeid, left_leaves, right_leaves, root)

    def print(self):
        print(
            "node_pos{}, attr:{}, th:{}, leaves:{}".format(
                self.node_pos, self.attribute, self.threshold, self.leaves_lists
            )
        )

    def add_leaves(self, treeid, nodeid, left_leaves, right_leaves, root=False):
        self.node_pos.append({"treeid": treeid, "nodeid": nodeid})
        if root:
            self.leaves_lists.append((left_leaves, right_leaves, "root"))
        else:
            self.leaves_lists.append((left_leaves, right_leaves))

    def add_grb_var(self, node_grb_var, leaf_grb_var_list):
        self.p_grb_var = node_grb_var
        self.l_grb_var_list = []
        for item in self.leaves_lists:
            left_leaf_grb_var = [leaf_grb_var_list[i] for i in item[0]]
            right_leaf_grb_var = [leaf_grb_var_list[i] for i in item[1]]
            if len(item) == 3:
                self.l_grb_var_list.append(
                    (left_leaf_grb_var, right_leaf_grb_var, "root")
                )
            else:
                self.l_grb_var_list.append((left_leaf_grb_var, right_leaf_grb_var))


class KantchelianAttack(object):
    def __init__(
        self,
        json_model,
        epsilon=None,
        order=np.inf,
        guard_val=GUARD_VAL,
        round_digits=ROUND_DIGITS,
        pos_json_input=None,
        neg_json_input=None,
        pred_threshold=0.0,
        verbose=False,
        n_threads=1,
    ):
        assert (
            epsilon is None or order == np.inf
        ), "feasibility epsilon can only be used with order inf"

        self.pred_threshold = pred_threshold
        self.epsilon = epsilon
        self.binary = (pos_json_input == None) or (neg_json_input == None)
        self.pos_json_input = pos_json_input
        self.neg_json_input = neg_json_input
        self.guard_val = guard_val
        self.round_digits = round_digits
        self.json_model = json_model
        self.order = order
        self.verbose = verbose
        self.n_threads = n_threads

        # two nodes with identical decision are merged in this list, their left and right leaves and in the list, third element of the tuple
        self.node_list = []
        self.leaf_v_list = []  # list of all leaf values
        self.leaf_pos_list = []  # list of leaves' position in xgboost model
        self.leaf_count = [0]  # total number of leaves in the first i trees
        node_check = (
            {}
        )  # track identical decision nodes. {(attr, th):<index in node_list>}

        def dfs(tree, treeid, root=False, neg=False):
            if "leaf" in tree.keys():
                if neg:
                    self.leaf_v_list.append(-tree["leaf"])
                else:
                    self.leaf_v_list.append(tree["leaf"])
                self.leaf_pos_list.append({"treeid": treeid, "nodeid": tree["nodeid"]})
                return [len(self.leaf_v_list) - 1]
            else:
                attribute, threshold, nodeid = (
                    tree["split"],
                    tree["split_condition"],
                    tree["nodeid"],
                )
                if type(attribute) == str:
                    attribute = int(attribute[1:])
                threshold = round(threshold, self.round_digits)
                # XGBoost can only offer precision up to 8 digits, however, minimum difference between two splits can be smaller than 1e-8
                # here rounding may be an option, but its hard to choose guard value after rounding
                # for example, if round to 1e-6, then guard value should be 5e-7, or otherwise may cause mistake
                # xgboost prediction has a precision of 1e-8, so when min_diff<1e-8, there is a precision problem
                # if we do not round, xgboost.predict may give wrong results due to precision, but manual predict on json file should always work
                left_subtree = None
                right_subtree = None
                for subtree in tree["children"]:
                    if subtree["nodeid"] == tree["yes"]:
                        left_subtree = subtree
                    if subtree["nodeid"] == tree["no"]:
                        right_subtree = subtree
                if left_subtree == None or right_subtree == None:
                    pprint.pprint(tree)
                    raise ValueError("should be a tree but one child is missing")
                left_leaves = dfs(left_subtree, treeid, False, neg)
                right_leaves = dfs(right_subtree, treeid, False, neg)
                if (attribute, threshold) not in node_check:
                    self.node_list.append(
                        node_wrapper(
                            treeid,
                            nodeid,
                            attribute,
                            threshold,
                            left_leaves,
                            right_leaves,
                            root,
                        )
                    )
                    node_check[(attribute, threshold)] = len(self.node_list) - 1
                else:
                    node_index = node_check[(attribute, threshold)]
                    self.node_list[node_index].add_leaves(
                        treeid, nodeid, left_leaves, right_leaves, root
                    )
                return left_leaves + right_leaves

        if self.binary:
            for i, tree in enumerate(self.json_model):
                dfs(tree, i, root=True)
                self.leaf_count.append(len(self.leaf_v_list))
            if len(self.json_model) + 1 != len(self.leaf_count):
                print("self.leaf_count:", self.leaf_count)
                raise ValueError("leaf count error")
        else:
            for i, tree in enumerate(self.pos_json_input):
                dfs(tree, i, root=True)
                self.leaf_count.append(len(self.leaf_v_list))
            for i, tree in enumerate(self.neg_json_input):
                dfs(tree, i + len(self.pos_json_input), root=True, neg=True)
                self.leaf_count.append(len(self.leaf_v_list))
            if len(self.pos_json_input) + len(self.neg_json_input) + 1 != len(
                self.leaf_count
            ):
                print("self.leaf_count:", self.leaf_count)
                raise ValueError("leaf count error")

        self.m = Model("attack")

        if not self.verbose:
            self.m.setParam(
                "OutputFlag", 0
            )  # suppress Gurobi output, gives a small speed-up and prevents huge logs

        self.m.setParam("Threads", self.n_threads)

        # Most datasets require a very low tolerance
        self.m.setParam("IntFeasTol", 1e-9)
        self.m.setParam("FeasibilityTol", 1e-9)

        self.P = self.m.addVars(len(self.node_list), vtype=GRB.BINARY, name="p")
        self.L = self.m.addVars(len(self.leaf_v_list), lb=0, ub=1, name="l")
        if epsilon:
            self.B = self.m.addVar(name="b", lb=0.0, ub=self.epsilon - 0.0001)
        elif self.order == np.inf:
            self.B = self.m.addVar(name="b")
        self.llist = [self.L[key] for key in range(len(self.L))]
        self.plist = [self.P[key] for key in range(len(self.P))]

        # p dictionary by attributes, {attr1:[(threshold1, gurobiVar1),(threshold2, gurobiVar2),...],attr2:[...]}
        self.pdict = {}
        for i, node in enumerate(self.node_list):
            node.add_grb_var(self.plist[i], self.llist)
            if node.attribute not in self.pdict:
                self.pdict[node.attribute] = [(node.threshold, self.plist[i])]
            else:
                self.pdict[node.attribute].append((node.threshold, self.plist[i]))

        # sort each feature list
        # add p constraints
        for key in self.pdict.keys():
            min_diff = 1000
            if len(self.pdict[key]) > 1:
                self.pdict[key].sort(key=lambda tup: tup[0])
                for i in range(len(self.pdict[key]) - 1):
                    self.m.addConstr(
                        self.pdict[key][i][1] <= self.pdict[key][i + 1][1],
                        name="p_consis_attr{}_{}th".format(key, i),
                    )
                    min_diff = min(
                        min_diff, self.pdict[key][i + 1][0] - self.pdict[key][i][0]
                    )

                if min_diff < 2 * self.guard_val:
                    self.guard_val = min_diff / 3
                    print(
                        "guard value too large, change to min_diff/3:", self.guard_val
                    )

        # all leaves sum up to 1
        for i in range(len(self.leaf_count) - 1):
            leaf_vars = [
                self.llist[j] for j in range(self.leaf_count[i], self.leaf_count[i + 1])
            ]
            self.m.addConstr(
                LinExpr([1] * (self.leaf_count[i + 1] - self.leaf_count[i]), leaf_vars)
                == 1,
                name="leaf_sum_one_for_tree{}".format(i),
            )

        # node leaves constraints
        for j in range(len(self.node_list)):
            p = self.plist[j]
            for k in range(len(self.node_list[j].leaves_lists)):
                left_l = [self.llist[i] for i in self.node_list[j].leaves_lists[k][0]]
                right_l = [self.llist[i] for i in self.node_list[j].leaves_lists[k][1]]
                if len(self.node_list[j].leaves_lists[k]) == 3:
                    self.m.addConstr(
                        LinExpr([1] * len(left_l), left_l) - p == 0,
                        name="p{}_root_left_{}".format(j, k),
                    )
                    self.m.addConstr(
                        LinExpr([1] * len(right_l), right_l) + p == 1,
                        name="p_{}_root_right_{}".format(j, k),
                    )
                else:
                    self.m.addConstr(
                        LinExpr([1] * len(left_l), left_l) - p <= 0,
                        name="p{}_left_{}".format(j, k),
                    )
                    self.m.addConstr(
                        LinExpr([1] * len(right_l), right_l) + p <= 1,
                        name="p{}_right_{}".format(j, k),
                    )
        self.m.update()

    def attack_feasible(self, sample, label):
        if self.binary:
            pred = (
                1 if self.check(sample, self.json_model) >= self.pred_threshold else 0
            )
        else:
            pred = (
                1
                if self.check(sample, self.pos_json_input)
                >= self.check(sample, self.neg_json_input)
                else 0
            )
        x = np.copy(sample)

        if pred != label:
            # Wrong prediction, no attack needed
            return True

        # model mislabel
        try:
            c = self.m.getConstrByName("mislabel")
            self.m.remove(c)
        except Exception:
            pass
        if (not self.binary) or label == 1:
            self.m.addConstr(
                LinExpr(self.leaf_v_list, self.llist)
                <= self.pred_threshold - self.guard_val,
                name="mislabel",
            )
        else:
            self.m.addConstr(
                LinExpr(self.leaf_v_list, self.llist)
                >= self.pred_threshold + self.guard_val,
                name="mislabel",
            )

        # Generate constraints for self.B, the l-infinity distance.
        for key in self.pdict.keys():
            if len(self.pdict[key]) == 0:
                raise ValueError("self.pdict list empty")
            axis = [-np.inf] + [item[0] for item in self.pdict[key]] + [np.inf]
            w = [0] * (len(self.pdict[key]) + 1)
            for i in range(len(axis) - 1, 0, -1):
                if x[key] < axis[i] and x[key] >= axis[i - 1]:
                    w[i - 1] = 0
                elif x[key] < axis[i] and x[key] < axis[i - 1]:
                    w[i - 1] = np.abs(x[key] - axis[i - 1])
                elif x[key] >= axis[i] and x[key] >= axis[i - 1]:
                    w[i - 1] = np.abs(x[key] - axis[i] + self.guard_val)
                else:
                    print("x[key]:", x[key])
                    print("axis:", axis)
                    print("axis[i]:{}, axis[i-1]:{}".format(axis[i], axis[i - 1]))
                    raise ValueError("wrong axis ordering")
            for i in range(len(w) - 1):
                w[i] -= w[i + 1]
            else:
                try:
                    c = self.m.getConstrByName("linf_constr_attr{}".format(key))
                    self.m.remove(c)
                except Exception:
                    pass
                self.m.addConstr(
                    LinExpr(w[:-1], [item[1] for item in self.pdict[key]]) + w[-1]
                    <= self.B,
                    name="linf_constr_attr{}".format(key),
                )

        self.m.setObjective(0, GRB.MINIMIZE)

        self.m.update()
        self.m.optimize()

        return not self.m.status == 3  # 3 -> infeasible -> no adv example -> False

    def optimal_adversarial_example(self, sample, label):
        if self.binary:
            pred = (
                1 if self.check(sample, self.json_model) >= self.pred_threshold else 0
            )
        else:
            pred = (
                1
                if self.check(sample, self.pos_json_input)
                >= self.check(sample, self.neg_json_input)
                else 0
            )
        x = np.copy(sample)

        if pred != label:
            # Wrong prediction, no attack needed
            return x

        # model mislabel
        # this is for binary
        try:
            c = self.m.getConstrByName("mislabel")
            self.m.remove(c)
        except Exception:
            pass
        if (not self.binary) or label == 1:
            self.m.addConstr(
                LinExpr(self.leaf_v_list, self.llist)
                <= self.pred_threshold - self.guard_val,
                name="mislabel",
            )
        else:
            self.m.addConstr(
                LinExpr(self.leaf_v_list, self.llist)
                >= self.pred_threshold + self.guard_val,
                name="mislabel",
            )

        if self.order == np.inf:
            rho = 1
        else:
            rho = self.order

        if self.order != np.inf:
            self.obj_coeff_list = []
            self.obj_var_list = []
            self.obj_c = 0
        # model objective
        for key in self.pdict.keys():
            if len(self.pdict[key]) == 0:
                raise ValueError("self.pdict list empty")
            axis = [-np.inf] + [item[0] for item in self.pdict[key]] + [np.inf]
            w = [0] * (len(self.pdict[key]) + 1)
            for i in range(len(axis) - 1, 0, -1):
                if x[key] < axis[i] and x[key] >= axis[i - 1]:
                    w[i - 1] = 0
                elif x[key] < axis[i] and x[key] < axis[i - 1]:
                    w[i - 1] = np.abs(x[key] - axis[i - 1]) ** rho
                elif x[key] >= axis[i] and x[key] >= axis[i - 1]:
                    w[i - 1] = np.abs(x[key] - axis[i] + self.guard_val) ** rho
                else:
                    print("x[key]:", x[key])
                    print("axis:", axis)
                    print("axis[i]:{}, axis[i-1]:{}".format(axis[i], axis[i - 1]))
                    raise ValueError("wrong axis ordering")
            for i in range(len(w) - 1):
                w[i] -= w[i + 1]
            if self.order != np.inf:
                self.obj_c += w[-1]
                self.obj_coeff_list += w[:-1]
                self.obj_var_list += [item[1] for item in self.pdict[key]]
            else:
                try:
                    c = self.m.getConstrByName("linf_constr_attr{}".format(key))
                    self.m.remove(c)
                except Exception:
                    pass
                self.m.addConstr(
                    LinExpr(w[:-1], [item[1] for item in self.pdict[key]]) + w[-1]
                    <= self.B,
                    name="linf_constr_attr{}".format(key),
                )

        if self.order != np.inf:
            self.m.setObjective(
                LinExpr(self.obj_coeff_list, self.obj_var_list) + self.obj_c,
                GRB.MINIMIZE,
            )
        else:
            self.m.setObjective(self.B, GRB.MINIMIZE)

        self.m.update()
        self.m.optimize()

        # If infeasible
        if self.m.status == 3:
            return None

        # Assert that the adversarial example causes a misclassification
        for key in self.pdict.keys():
            for node in self.pdict[key]:
                if node[1].x > 0.5 and x[key] >= node[0]:
                    x[key] = node[0] - self.guard_val
                if node[1].x <= 0.5 and x[key] < node[0]:
                    x[key] = node[0] + self.guard_val

        if self.binary:
            pred = 1 if self.check(x, self.json_model) >= self.pred_threshold else 0
        else:
            pos_value = self.check(x, self.pos_json_input)
            neg_value = self.check(x, self.neg_json_input)
            pred = 1 if pos_value >= neg_value else 0

        if pred == label and self.verbose:
            print("!" * 50)
            print("MILP result did not cause a misclassification!")
            print("!" * 50)

        return x

    def check(self, x, json_file):
        # Due to XGBoost precision issues, some attacks may not succeed if tested using model.predict.
        # We manually run the tree on the json file here to make sure those attacks are actually successful.
        leaf_values = []
        for item in json_file:
            tree = item.copy()
            while "leaf" not in tree.keys():
                attribute, threshold, nodeid = (
                    tree["split"],
                    tree["split_condition"],
                    tree["nodeid"],
                )
                if type(attribute) == str:
                    attribute = int(attribute[1:])
                if x[attribute] < threshold:
                    if tree["children"][0]["nodeid"] == tree["yes"]:
                        tree = tree["children"][0].copy()
                    elif tree["children"][1]["nodeid"] == tree["yes"]:
                        tree = tree["children"][1].copy()
                    else:
                        pprint.pprint(tree)
                        print("x[attribute]:", x[attribute])
                        raise ValueError("child not found")
                else:
                    if tree["children"][0]["nodeid"] == tree["no"]:
                        tree = tree["children"][0].copy()
                    elif tree["children"][1]["nodeid"] == tree["no"]:
                        tree = tree["children"][1].copy()
                    else:
                        pprint.pprint(tree)
                        print("x[attribute]:", x[attribute])
                        raise ValueError("child not found")
            leaf_values.append(tree["leaf"])
        manual_res = np.sum(leaf_values)
        return manual_res


class KantchelianAttackMultiClass(object):
    def __init__(
        self,
        json_model,
        n_classes,
        order=np.inf,
        epsilon=None,
        guard_val=GUARD_VAL,
        round_digits=ROUND_DIGITS,
        pred_threshold=0.0,
        low_memory=False,
        verbose=False,
        n_threads=1,
    ):
        if n_classes <= 2:
            raise ValueError("multiclass attack must be used when number of class > 2")

        assert (
            epsilon is None or order == np.inf
        ), "feasibility epsilon can only be used with order inf"

        self.n_classes = n_classes
        self.order = order
        self.epsilon = epsilon
        self.guard_val = guard_val
        self.round_digits = round_digits
        self.pred_threshold = pred_threshold
        self.low_memory = low_memory
        self.verbose = verbose
        self.n_threads = n_threads

        # Create all attacker models, this takes quadratic space in terms
        # of n_classes, but speeds up attacks for many samples.
        self.one_vs_all_models = [[] for _ in range(self.n_classes)]
        for i, json_tree in enumerate(json_model):
            self.one_vs_all_models[i % n_classes].append(json_tree)

        if not low_memory:
            self.__create_cached_attackers()

    def __create_cached_attackers(self):
        self.binary_attackers = []
        for class_label in range(self.n_classes):
            attackers = []
            for other_label in range(self.n_classes):
                if class_label == other_label:
                    attackers.append(None)
                    continue

                attacker = KantchelianAttack(
                    None,
                    epsilon=self.epsilon,
                    order=self.order,
                    guard_val=self.guard_val,
                    round_digits=self.round_digits,
                    pred_threshold=self.pred_threshold,
                    verbose=self.verbose,
                    n_threads=self.n_threads,
                    pos_json_input=self.one_vs_all_models[class_label],
                    neg_json_input=self.one_vs_all_models[other_label],
                )

                attackers.append(attacker)
            self.binary_attackers.append(attackers)
        return self.binary_attackers

    def optimal_adversarial_example(self, sample, label):
        best_distance = float("inf")
        best_adv_example = None

        for other_label in range(self.n_classes):
            if other_label == label:
                continue

            # Create new attacker or use a cached attacker
            if self.low_memory:
                attacker = KantchelianAttack(
                    None,
                    epsilon=self.epsilon,
                    order=self.order,
                    guard_val=self.guard_val,
                    round_digits=self.round_digits,
                    pred_threshold=self.pred_threshold,
                    verbose=self.verbose,
                    n_threads=self.n_threads,
                    pos_json_input=self.one_vs_all_models[label],
                    neg_json_input=self.one_vs_all_models[other_label],
                )
            else:
                attacker = self.binary_attackers[label][other_label]

            # Generate adversarial example on this binary attacker
            adv_example = attacker.optimal_adversarial_example(sample, 1)

            # If this binary attacker example was better than the previous ones, keep it
            if adv_example is not None:
                distance = np.linalg.norm(sample - adv_example, ord=self.order)
                if distance < best_distance:
                    best_adv_example = adv_example
                    best_distance = distance

        if best_adv_example is None:
            raise Exception(
                "No adversarial example found, does your model predict a constant value?"
            )

        return best_adv_example

    def attack_feasible(self, sample, label):
        for other_label in range(self.n_classes):
            if other_label == label:
                continue

            # Create new attacker or use a cached attacker
            if self.low_memory:
                attacker = KantchelianAttack(
                    None,
                    epsilon=self.epsilon,
                    order=self.order,
                    guard_val=self.guard_val,
                    round_digits=self.round_digits,
                    pred_threshold=self.pred_threshold,
                    verbose=self.verbose,
                    n_threads=self.n_threads,
                    pos_json_input=self.one_vs_all_models[label],
                    neg_json_input=self.one_vs_all_models[other_label],
                )
            else:
                attacker = self.binary_attackers[label][other_label]

            # Check if the binary attacker can create an adversarial example
            if attacker.attack_feasible(sample, 1):
                return True

        return False


DEFAULT_OPTIONS = {
    "epsilon": None,
    "guard_val": GUARD_VAL,
    "round_digits": ROUND_DIGITS,
    "pred_threshold": 0.0,
    "order": np.inf,
    "low_memory": False,
    "verbose": False,
    "n_threads": 1,
}


class KantchelianAttackWrapper(AttackWrapper):
    def __init__(self, json_model, n_classes):
        self.json_model = json_model
        self.n_classes = n_classes

    def __get_attacker(self, order, options):
        if self.n_classes == 2:
            attack = KantchelianAttack(
                self.json_model,
                order=order,
                epsilon=options["epsilon"],
                guard_val=options["guard_val"],
                round_digits=options["round_digits"],
                pred_threshold=options["pred_threshold"],
                verbose=options["verbose"],
                n_threads=options["n_threads"],
            )
        else:
            attack = KantchelianAttackMultiClass(
                self.json_model,
                self.n_classes,
                order=order,
                epsilon=options["epsilon"],
                guard_val=options["guard_val"],
                round_digits=options["round_digits"],
                pred_threshold=options["pred_threshold"],
                low_memory=options["low_memory"],
                verbose=options["verbose"],
                n_threads=options["n_threads"],
            )
        return attack

    def attack_feasibility(self, X, y, order, epsilon, options={}):
        default_options = DEFAULT_OPTIONS.copy()
        default_options.update(options)
        options = default_options
        options["epsilon"] = epsilon

        attack = self.__get_attacker(order, options)

        attack_feasible = []
        start_time = time.time()

        for sample, label in tqdm(zip(X, y), total=X.shape[0]):
            attack_feasible.append(attack.attack_feasible(sample, label))

        total_time = time.time() - start_time

        if options["verbose"]:
            print("Total time:", total_time)
            print("Avg time per instance:", total_time / len(X))

        return np.array(attack_feasible)

    def adversarial_examples(self, X, y, order, options={}):
        default_options = DEFAULT_OPTIONS.copy()
        default_options.update(options)
        options = default_options

        attack = self.__get_attacker(order, options)

        start_time = time.time()

        X_adv = []
        for sample, label in tqdm(zip(X, y), total=X.shape[0]):
            optimal_example = attack.optimal_adversarial_example(sample, label)
            X_adv.append(optimal_example)

        total_time = time.time() - start_time

        if options["verbose"]:
            print("Total time:", total_time)
            print("Avg time per instance:", total_time / len(X))

        return np.array(X_adv)



import numpy as np
import csv
import scipy.io
from tensorflow.keras.datasets import (
    mnist as mnist_keras,
    fashion_mnist as fashion_mnist_keras,
    cifar10 as cifar10_keras,
)

from ..datasets import load_mnist

from sklearn.model_selection import train_test_split
from sklearn.preprocessing import MinMaxScaler

data_dir = "/home/maksym/boost/data/"


def split_train_test(X_all, y_all, frac_train):
    """
    The first X% of X_all, y_all become the training set, the rest (1-X)% become the test set.
    Note that this assumes that the samples are already shuffled or if not (e.g. if we were to split MNIST) that
    this behavior is intended.
    """
    num_total = X_all.shape[0]
    num_train = int(frac_train * num_total)

    X_train, y_train = X_all[:num_train], y_all[:num_train]
    X_test, y_test = X_all[num_train:], y_all[num_train:]

    return X_train, y_train, X_test, y_test


def normalize_per_feature_0_1(X_train, X_test):
    """
    We are not allowed to touch the test data, thus we do the normalization just based on the training data.
    """
    X_train_max = X_train.max(axis=0, keepdims=True)
    X_train_min = X_train.min(axis=0, keepdims=True)
    X_train = (X_train - X_train_min) / (X_train_max - X_train_min)
    X_test = (X_test - X_train_min) / (X_train_max - X_train_min)
    return X_train, X_test


def split_train_validation(X_train_orig, y_train_orig, frac_valid, shuffle=True):
    num_total = X_train_orig.shape[0]
    n_valid = int(frac_valid * num_total)
    idx = np.random.permutation(num_total) if shuffle else np.arange(num_total)
    if shuffle:
        X_valid, y_valid = X_train_orig[idx][:n_valid], y_train_orig[idx][:n_valid]
        X_train, y_train = X_train_orig[idx][n_valid:], y_train_orig[idx][n_valid:]
    else:
        # If no shuffle, then one has to ensure that the classes are balanced
        idx_valid, idx_train = [], []
        for cls in np.unique(y_train_orig):
            indices_cls = np.where(y_train_orig == cls)[0]
            proportion_cls = len(indices_cls) / num_total
            n_class_balanced_valid = int(proportion_cls * n_valid)
            idx_valid.extend(list(indices_cls[:n_class_balanced_valid]))
            idx_train.extend(list(indices_cls[n_class_balanced_valid:]))
        idx_valid, idx_train = np.array(idx_valid), np.array(idx_train)
        X_valid, y_valid = X_train_orig[idx_valid], y_train_orig[idx_valid]
        X_train, y_train = X_train_orig[idx_train], y_train_orig[idx_train]
    return X_train, y_train, X_valid, y_valid


def binary_from_multiclass(X_train, y_train, X_test, y_test, classes):
    classes = np.array(classes)  # for indexing only arrays work, not lists

    idx_train1, idx_train2 = y_train == classes[0], y_train == classes[1]
    idx_test1, idx_test2 = y_test == classes[0], y_test == classes[1]
    X_train, X_test = X_train[idx_train1 + idx_train2], X_test[idx_test1 + idx_test2]

    y_train = idx_train1 * 1 + idx_train2 * -1
    y_test = idx_test1 * 1 + idx_test2 * -1
    y_train, y_test = y_train[idx_train1 + idx_train2], y_test[idx_test1 + idx_test2]

    return X_train, y_train, X_test, y_test


def transform_labels_one_vs_all(y_train_orig, y_valid_orig, y_test_orig):
    n_cls = int(y_train_orig.max()) + 1
    if n_cls == 2:
        return y_train_orig[None, :], y_valid_orig[None, :], y_test_orig[None, :]

    labels = np.unique(y_train_orig)
    n_cls = len(labels)
    n_train, n_valid, n_test = (
        y_train_orig.shape[0],
        y_valid_orig.shape[0],
        y_test_orig.shape[0],
    )
    y_train, y_valid, y_test = (
        np.zeros([n_cls, n_train]),
        np.zeros([n_cls, n_valid]),
        np.zeros([n_cls, n_test]),
    )
    for i_cls in range(n_cls):
        # convert from False/True to -1/1 compatible with One-vs-All formulation
        y_train[i_cls] = 2 * (y_train_orig == i_cls) - 1
        y_valid[i_cls] = 2 * (y_valid_orig == i_cls) - 1
        y_test[i_cls] = 2 * (y_test_orig == i_cls) - 1
    return y_train, y_valid, y_test


def toy_2d_stumps():
    X = np.array(
        [
            [0.38, 0.75],
            [0.50, 0.93],
            [0.05, 0.70],
            [0.30, 0.90],
            [0.15, 0.80],
            # [0.15, 1.0], [0.125, 0.75], [0.1, 0.85], [0.045, 0.22], [0.725, 0.955],  # small margin
            # [0.15, 1.0], [0.125, 0.75], [0.1, 0.85], [0.075, 0.2], [0.775, 0.925],  # small margin
            [0.15, 1.0],
            [0.125, 0.5],
            [0.1, 0.85],
            [0.02, 0.25],
            [0.775, 0.975],
            [0.05, 0.05],
            [0.2, 0.1],
            [0.4, 0.075],
            [0.6, 0.22],
            [0.8, 0.1],
            [0.95, 0.05],
            [0.9, 0.2],
            [0.925, 0.4],
            [0.79, 0.6],
            [0.81, 0.8],
        ]
    )
    y = np.array([-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    eps_dataset = 0.075
    return X, y, eps_dataset


def toy_2d_trees():
    X = np.array(
        [
            [0.38, 0.75],
            [0.50, 0.93],
            [0.05, 0.70],
            [0.30, 0.90],
            [0.15, 0.80],
            [0.75, 0.38],
            [0.95, 0.48],
            [0.70, 0.05],
            [0.65, 0.30],
            [0.80, 0.30],
            [0.05, 0.1],
            [0.35, 0.1],
            [0.45, 0.075],
            [0.3, 0.2],
            [0.25, 0.1],
            [0.95, 0.65],
            [0.7, 0.9],
            [0.925, 0.7],
            [0.79, 0.55],
            [0.81, 0.8],
        ]
    )
    y = np.array([-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1])
    eps_dataset = 0.075
    return X, y, eps_dataset


def toy_2d_xor():
    X = np.array([[0.05, 0.05], [0.95, 0.95], [0.05, 0.95], [0.95, 0.05]])
    y = np.array([-1, -1, 1, 1])
    eps_dataset = 0.15
    return X, y, eps_dataset


def toy_2d_wong():
    # random points at least 2r apart
    m = 12
    # seed=10 illustrates that by default the margin can be easily close to 0
    # both plain and robust model have 0 train error, but the robust model additionally enforces a large margin
    np.random.seed(10)
    x = [np.random.uniform(size=2)]
    r = 0.16
    while len(x) < m:
        p = np.random.uniform(size=2)
        if min(np.abs(p - a).sum() for a in x) > 2 * r:
            x.append(p)
    eps_dataset = r / 2

    X = np.array(x)
    y = np.sign(np.random.uniform(-0.5, 0.5, size=m))
    return X, y, eps_dataset


def breast_cancer():
    """
    Taken from the UCI repository:
    http://archive.ics.uci.edu/ml/datasets/breast+cancer+wisconsin+%28diagnostic%29 file: breast-cancer-wisconsin.data

    After filtering the points with missing data, we have exactly the same as Chen et al, 2019
    train: 546x10, test: 137x10
    """
    eps_dataset = 0.3  # same as in Chen et al, 2019, worked well for them
    path = data_dir + "breast_cancer/breast-cancer-wisconsin.data"

    lst = []
    for line in csv.reader(open(path, "r").readlines()):
        if "?" not in line:
            lst.append(line)
    data_arr = np.array(lst, dtype=int)

    X_all, y_all = data_arr[:, :10], data_arr[:, 10]
    y_all[y_all == 2], y_all[y_all == 4] = -1, 1  # from 2, 4 to -1, 1

    X_train, y_train, X_test, y_test = split_train_test(X_all, y_all, frac_train=0.8)
    X_train, X_test = normalize_per_feature_0_1(X_train, X_test)

    return X_train, y_train, X_test, y_test, eps_dataset


def diabetes():
    """
    Taken from Kaggle:
    https://www.kaggle.com/uciml/pima-indians-diabetes-database file: diabetes.csv

    train: 614x8, test: 154x8
    """
    eps_dataset = 0.05  # Chen et al, 2019 used 0.2, but it was too high
    path = data_dir + "diabetes/diabetes.csv"
    data_arr = np.loadtxt(path, delimiter=",", skiprows=1)  # loaded as float64

    X_all, y_all = data_arr[:, :8], data_arr[:, 8]
    y_all[y_all == 0], y_all[y_all == 1] = -1, 1  # from 0, 1 to -1, 1

    X_train, y_train, X_test, y_test = split_train_test(X_all, y_all, frac_train=0.8)
    X_train, X_test = normalize_per_feature_0_1(X_train, X_test)

    return X_train, y_train, X_test, y_test, eps_dataset


def ijcnn1():
    """
    Taken from LIBSVM data repository:
    https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary.html

    train: 49990x22, test: 91701x22
    note: imbalanced classes (-1: 90.3% vs 1: 9.7%)
    """
    eps_dataset = 0.01  # Chen et al, 2019 used 0.1, but it was too high
    folder = data_dir + "ijcnn1/"
    path_train, path_val, path_test = (
        folder + "ijcnn1.tr",
        folder + "ijcnn1.val",
        folder + "ijcnn1.t",
    )

    num_train, num_test, dim = 49990, 91701, 22
    X_train = np.zeros((num_train, dim))
    y_train = np.zeros(num_train)
    num_train_orig = 0
    for i, line in enumerate(open(path_train, "r").readlines()):
        y_train[i] = int(float(line.split(" ")[0]))  # -1 or 1
        for s in line.split(" ")[1:]:
            coord_str, val_str = s.replace("\n", "").split(":")
            coord, val = (
                int(coord_str) - 1,
                float(val_str),
            )  # -1 is needed to have pythonic numeration from 0
            X_train[i, coord] = val
        num_train_orig += 1

    num_val_orig = 0
    for i, line in enumerate(open(path_val, "r").readlines()):
        y_train[num_train_orig + i] = int(float(line.split(" ")[0]))  # -1 or 1
        for s in line.split(" ")[1:]:
            coord_str, val_str = s.replace("\n", "").split(":")
            coord, val = int(coord_str) - 1, float(val_str)
            X_train[num_train_orig + i, coord] = val
        num_val_orig += 1

    assert (
        num_train == num_train_orig + num_val_orig
    )  # Check that we have not introduced extra zero rows

    X_test = np.zeros((num_test, dim))
    y_test = np.zeros(num_test)
    num_test_orig = 0
    for i, line in enumerate(open(path_test, "r").readlines()):
        y_test[i] = int(float(line.split(" ")[0]))  # -1 or 1
        for s in line.split(" ")[1:]:
            coord_str, val_str = s.replace("\n", "").split(":")
            coord, val = int(coord_str) - 1, float(val_str)
            X_test[i, coord] = val
        num_test_orig += 1

    assert num_test == num_test_orig

    X_train, X_test = normalize_per_feature_0_1(X_train, X_test)

    return X_train, y_train, X_test, y_test, eps_dataset


def cod_rna():
    """
    Taken from LIBSVM data repository:
    https://www.csie.ntu.edu.tw/~cjlin/libsvmtools/datasets/binary.html

    train: 59535x8, test: 271617x8
    """
    eps_dataset = 0.025  # Chen et al, 2019 used 0.2, but it was too high
    folder = data_dir + "cod_rna/"
    path_train, path_test = folder + "cod-rna.tr", folder + "cod-rna.t"

    num_train, num_test, dim = 59535, 271617, 8
    X_train = np.zeros((num_train, dim))
    y_train = np.zeros(num_train)
    num_train_orig = 0
    for i, line in enumerate(open(path_train, "r").readlines()):
        y_train[i] = int(float(line.split(" ")[0]))  # -1 or 1
        for s in line.split(" ")[1:]:
            coord_str, val_str = s.replace("\n", "").split(":")
            coord, val = (
                int(coord_str) - 1,
                float(val_str),
            )  # -1 is needed to have pythonic numeration from 0
            X_train[i, coord] = val
        num_train_orig += 1

    assert (
        num_train == num_train_orig
    )  # Check that we have not introduced extra zero rows

    X_test = np.zeros((num_test, dim))
    y_test = np.zeros(num_test)
    num_test_orig = 0
    for i, line in enumerate(open(path_test, "r").readlines()):
        y_test[i] = int(float(line.split(" ")[0]))  # -1 or 1
        for s in line.split(" ")[1:]:
            coord_str, val_str = s.replace("\n", "").split(":")
            coord, val = int(coord_str) - 1, float(val_str)
            X_test[i, coord] = val
        num_test_orig += 1

    assert num_test == num_test_orig

    X_train, X_test = normalize_per_feature_0_1(X_train, X_test)

    # n_test_final = 10000  # take 10k test examples instead of all 270k
    n_test_final = num_test
    idx = np.random.permutation(num_test)[:n_test_final]
    X_test, y_test = X_test[idx], y_test[idx]
    return X_train, y_train, X_test, y_test, eps_dataset


def mnist_1_5():
    """
    train: (12163, 784), test: (2027, 784)
    """
    eps_dataset = 0.3
    classes = [1, 5]  # 2 is 1, 6 is -1 in the binary classification scheme

    (X_train, y_train), (X_test, y_test) = mnist_keras.load_data()
    X_train, X_test = (
        X_train.astype(np.float64) / 255.0,
        X_test.astype(np.float64) / 255.0,
    )
    X_train = np.reshape(X_train, [X_train.shape[0], -1])
    X_test = np.reshape(X_test, [X_test.shape[0], -1])

    X_train, y_train, X_test, y_test = binary_from_multiclass(
        X_train, y_train, X_test, y_test, classes
    )
    return X_train, y_train, X_test, y_test, eps_dataset


def mnist_2_6():
    """
    train: (11876, 784), test: (1990, 784)
    """
    eps_dataset = 0.3
    classes = [2, 6]  # 2 is 1, 6 is -1 in the binary classification scheme

    (X_train, y_train), (X_test, y_test) = mnist_keras.load_data()
    X_train, X_test = (
        X_train.astype(np.float64) / 255.0,
        X_test.astype(np.float64) / 255.0,
    )
    X_train = np.reshape(X_train, [X_train.shape[0], -1])
    X_test = np.reshape(X_test, [X_test.shape[0], -1])

    X_train, y_train, X_test, y_test = binary_from_multiclass(
        X_train, y_train, X_test, y_test, classes
    )
    return X_train, y_train, X_test, y_test, eps_dataset


# def mnist_2_6():
#     """
#     train: (11876, 784), test: (1990, 784)
#     """
#     eps_dataset = 0.3

#     X, y = load_mnist()[1:3]
#     X = MinMaxScaler().fit_transform(X)

#     X = X[(y == 2) | (y == 6)]
#     y = y[(y == 2) | (y == 6)]
#     y = np.where(y == 2, 1, -1)

#     X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.3, random_state=1)

#     return X_train, y_train, X_test, y_test, eps_dataset


def mnist():
    """
    train: (60000, 784), test: (10000, 784)
    """
    eps_dataset = 0.3

    (X_train, y_train), (X_test, y_test) = mnist_keras.load_data()
    X_train, X_test = (
        X_train.astype(np.float64) / 255.0,
        X_test.astype(np.float64) / 255.0,
    )
    X_train = np.reshape(X_train, [X_train.shape[0], -1])
    X_test = np.reshape(X_test, [X_test.shape[0], -1])

    return X_train, y_train, X_test, y_test, eps_dataset


def cifar10():
    """
    train: (60000, 3072), test: (10000, 3072)
    """
    eps_dataset = 8 / 255

    (X_train, y_train), (X_test, y_test) = cifar10_keras.load_data()
    X_train, X_test = (
        X_train.astype(np.float64) / 255.0,
        X_test.astype(np.float64) / 255.0,
    )
    X_train = np.reshape(X_train, [X_train.shape[0], -1])
    X_test = np.reshape(X_test, [X_test.shape[0], -1])

    y_train, y_test = y_train.flatten(), y_test.flatten()

    return X_train, y_train, X_test, y_test, eps_dataset


def fmnist_sandal_sneaker():
    """
    Classes:
    0	T-shirt/top
    1	Trouser
    2	Pullover
    3	Dress
    4	Coat
    5	Sandal
    6	Shirt
    7	Sneaker
    8	Bag
    9	Ankle boot

    train: (12000, 784), test: (2000, 784)
    """
    eps_dataset = 0.1
    classes = [5, 7]  # 5 is 1, 7 is -1 in the binary classification scheme

    (X_train, y_train), (X_test, y_test) = fashion_mnist_keras.load_data()
    X_train, X_test = (
        X_train.astype(np.float64) / 255.0,
        X_test.astype(np.float64) / 255.0,
    )
    X_train = np.reshape(X_train, [X_train.shape[0], -1])
    X_test = np.reshape(X_test, [X_test.shape[0], -1])

    X_train, y_train, X_test, y_test = binary_from_multiclass(
        X_train, y_train, X_test, y_test, classes
    )
    return X_train, y_train, X_test, y_test, eps_dataset


def fmnist():
    """
    train: (60000, 784), test: (10000, 784)
    """
    eps_dataset = 0.1

    (X_train, y_train), (X_test, y_test) = fashion_mnist_keras.load_data()
    X_train, X_test = (
        X_train.astype(np.float64) / 255.0,
        X_test.astype(np.float64) / 255.0,
    )
    X_train = np.reshape(X_train, [X_train.shape[0], -1])
    X_test = np.reshape(X_test, [X_test.shape[0], -1])

    return X_train, y_train, X_test, y_test, eps_dataset


def gts_100_roadworks():
    """
    the class ids can be checked in the original data folders, for example:
    1: speed 30, 4: speed 70, 7: speed 100, 8: speed 120, 18: warning, 25: roadworks

    train: (2940, 3072), test: (930, 3072)
    """
    eps_dataset = 8 / 255  # following Madry et al, 2017 for cifar10
    classes = [7, 25]

    # Originally, all pixels values are uint8 values in [0, 255]
    train = scipy.io.loadmat(data_dir + "gts/gts_int_train.mat")
    test = scipy.io.loadmat(data_dir + "gts/gts_int_test.mat")
    X_train, y_train, X_test, y_test = (
        train["images"],
        train["labels"],
        test["images"],
        test["labels"],
    )
    X_train, X_test = (
        X_train.reshape(X_train.shape[0], -1),
        X_test.reshape(X_test.shape[0], -1),
    )
    X_train, X_test = X_train / 255.0, X_test / 255.0
    y_train, y_test = y_train[0], y_test[0]  # get rid of the extra dimension

    X_train, y_train, X_test, y_test = binary_from_multiclass(
        X_train, y_train, X_test, y_test, classes
    )
    return X_train, y_train, X_test, y_test, eps_dataset


def gts_30_70():
    """
    the class ids can be checked in the original data folders, for example:
    1: speed 30, 4: speed 70, 7: speed 100, 8: speed 120, 18: warning, 25: roadworks

    train: 4200x3072, test: 1380x3072
    """
    eps_dataset = 8 / 255  # following Madry et al, 2017 for cifar10
    classes = [1, 4]

    # Originally, all pixels values are uint8 values in [0, 255]
    train = scipy.io.loadmat(data_dir + "gts/gts_int_train.mat")
    test = scipy.io.loadmat(data_dir + "gts/gts_int_test.mat")
    X_train, y_train, X_test, y_test = (
        train["images"],
        train["labels"],
        test["images"],
        test["labels"],
    )
    X_train, X_test = (
        X_train.reshape(X_train.shape[0], -1),
        X_test.reshape(X_test.shape[0], -1),
    )
    X_train, X_test = X_train / 255.0, X_test / 255.0
    y_train, y_test = y_train[0], y_test[0]  # get rid of the extra dimension

    X_train, y_train, X_test, y_test = binary_from_multiclass(
        X_train, y_train, X_test, y_test, classes
    )
    return X_train, y_train, X_test, y_test, eps_dataset


def har():
    """
    Human activity recognition dataset from https://archive.ics.uci.edu/ml/datasets/human+activity+recognition+using+smartphones
    Note: Wong and Kolter, ICML 2018 used eps=0.05, but the data points were from -1 to 1.
    We use equivalently eps=0.025, but data points from 0 to 1.

    The labels are in {0, 1, 2, 3, 4, 5}.

    train: (7352, 561), test: (2947, 561), n classes: 6.
    """
    eps_dataset = 0.025
    path_train, path_test = data_dir + "har/train/", data_dir + "har/test/"
    X_train, X_test = (
        np.loadtxt(path_train + "X_train.txt"),
        np.loadtxt(path_test + "X_test.txt"),
    )
    y_train, y_test = (
        np.loadtxt(path_train + "y_train.txt"),
        np.loadtxt(path_test + "y_test.txt"),
    )
    y_train, y_test = y_train - 1, y_test - 1  # make the class numeration start from 0
    X_train, X_test = (X_train + 1) / 2, (X_test + 1) / 2  # from [-1, 1] to [0, 1]
    return X_train, y_train, X_test, y_test, eps_dataset


def convert_to_float32(X):
    return X.astype(np.float32)


def random_crop(image, n_crop):
    h, w, _ = image.shape
    top = np.random.randint(0, n_crop)
    left = np.random.randint(0, n_crop)
    bottom = h - (n_crop - top)
    right = w - (n_crop - left)
    image = image[top:bottom, left:right, :]
    return image


def horizontal_flip(images, prob=0.5):
    if np.random.rand() < prob:
        images = images[:, :, ::-1, :]
    return images


def data_augment(X, dataset):
    num, dim = X.shape
    img_shape = datasets_img_shapes[dataset]
    X_img = np.reshape(np.copy(X), [num, *img_shape])
    if len(img_shape) == 2:  # introduce a fake last dimension for grayscale datasets
        X_img = X_img[:, :, :, None]

    n_crop = 2
    X_img_pad = np.pad(
        X_img,
        [(0, 0), (n_crop // 2, n_crop // 2), (n_crop // 2, n_crop // 2), (0, 0)],
        "constant",
        constant_values=0,
    )  # zero padding
    for i in range(num):
        X_img[i] = random_crop(
            X_img_pad[i], n_crop=n_crop
        )  # up to `n_crop` pixels are cropped

    if dataset in ["cifar10"]:
        X_img = horizontal_flip(X_img)

    return np.reshape(X_img, [num, dim])


def crop_batch(X_img, n_h, n_w, n_crop):
    _, h, w, _ = X_img.shape
    bottom, right = h - (n_crop - n_h), w - (n_crop - n_w)
    return X_img[:, n_h:bottom, n_w:right, :]


def extend_dataset(X, dataset):
    num, dim = X.shape
    img_shape = datasets_img_shapes[dataset]
    X_img = np.reshape(np.copy(X), [num, *img_shape])
    if len(img_shape) == 2:  # introduce a fake last dimension for grayscale datasets
        X_img = X_img[:, :, :, None]

    n_crop = 2
    X_img_pad = np.pad(
        X_img,
        [(0, 0), (n_crop // 2, n_crop // 2), (n_crop // 2, n_crop // 2), (0, 0)],
        "constant",
        constant_values=0,
    )

    # Note: (1, 1) is the original image
    X_img_l = crop_batch(X_img_pad, 1, 0, n_crop)
    X_img_r = crop_batch(X_img_pad, 1, 2, n_crop)
    X_img_t = crop_batch(X_img_pad, 0, 1, n_crop)
    X_img_b = crop_batch(X_img_pad, 2, 1, n_crop)

    X_img_extended = np.vstack([X_img, X_img_l, X_img_r, X_img_t, X_img_b])

    # if dataset in ['cifar10']:  # would lead to 10x expansion of the training data - might be too comp. expensive
    #     X_img_horiz_flip = X_img_extended[:, :, ::-1, :]
    #     X_img_extended = np.vstack([X_img_extended, X_img_horiz_flip])

    X_final = np.reshape(X_img_extended, [-1, dim])
    return X_final


all_datasets_dict = {
    "toy_2d_stumps": toy_2d_stumps,
    "toy_2d_trees": toy_2d_trees,
    "toy_2d_xor": toy_2d_xor,
    "toy_2d_wong": toy_2d_wong,
    "breast_cancer": breast_cancer,
    "diabetes": diabetes,
    "ijcnn1": ijcnn1,
    "cod_rna": cod_rna,
    "mnist_1_5": mnist_1_5,
    "mnist_2_6": mnist_2_6,
    "fmnist_sandal_sneaker": fmnist_sandal_sneaker,
    "gts_100_roadworks": gts_100_roadworks,
    "gts_30_70": gts_30_70,
    "har": har,
    "mnist": mnist,
    "fmnist": fmnist,
    "cifar10": cifar10,
}
dataset_names_dict = {
    "toy_2d_stumps": "toy_2d_stumps",
    "toy_2d_trees": "toy_2d_trees",
    "toy_2d_xor": "toy_2d_xor",
    "toy_2d_wong": "toy_2d_wong",
    "breast_cancer": "breast-cancer",
    "diabetes": "diabetes",
    "ijcnn1": "IJCNN1",
    "cod_rna": "cod-rna",
    "mnist_1_5": "MNIST 1-5",
    "mnist_2_6": "MNIST 2-6",
    "fmnist_sandal_sneaker": "FMNIST shoes",
    "gts_100_roadworks": "GTS 100-rw",
    "gts_30_70": "GTS 30-70",
    "har": "har",
    "mnist": "mnist",
    "fmnist": "fmnist",
    "cifar10": "cifar10",
}
datasets_img_shapes = {
    "mnist_1_5": (28, 28),
    "mnist_2_6": (28, 28),
    "mnist": (28, 28),
    "fmnist": (28, 28),
    "fmnist_sandal_sneaker": (28, 28),
    "gts_100_roadworks": (32, 32, 3),
    "gts_30_70": (32, 32, 3),
    "cifar10": (32, 32, 3),
}
datasets_feature_names = {
    "breast_cancer": [
        "radius",
        "texture",
        "perimeter",
        "area",
        "smoothness",
        "compactness",
        "concavity",
        "concave points",
        "symmetry",
        "fractal dimension",
    ],
    "diabetes": [
        "# pregnancies",
        "glucose",
        "blood pressure",
        "skin thickness",
        "insulin",
        "body mass index",
        "diabetes pedigree",
        "age",
    ],
    "cod_rna": [
        "Dynalign score",
        "shorter seq. length",
        "A freq. of seq. 1",
        "U freq. of seq. 1",
        "C freq. of seq. 1",
        "A freq. of seq. 2",
        "U freq. of seq. 2",
        "C freq. of seq. 2",
    ],
}



import os
import numpy as np
import glob
from numba import jit


class Logger:
    def __init__(self, path):
        self.path = path
        if path != "":
            folder = "/".join(path.split("/")[:-1])
            if not os.path.exists(folder):
                os.makedirs(folder)

    def print(self, message):
        print(message)
        if self.path != "":
            with open(self.path, "a") as f:
                f.write(message + "\n")
                f.flush()


def get_contiguous_indices(indices_opt_init):
    # needed when indices_opt_init are not contiguous (e.g. mnist_2_6: coord=613 has [12, 13, 15, 16, 18])
    running_diffs = indices_opt_init[1:] - indices_opt_init[:-1]  # [1, 2, 1, 2]
    where_change_contiguous_regions = np.where(running_diffs != 1)[
        0
    ]  # find the last contiguous index - 1

    if len(where_change_contiguous_regions) > 0:
        last_el_first_contiguous_region = where_change_contiguous_regions[0]
    elif (
        np.sum(running_diffs != np.ones(len(running_diffs))) == 0
    ):  # if all are optimal (mnist_2_6: coord=72)
        last_el_first_contiguous_region = len(running_diffs)
    elif (
        len(indices_opt_init) == 1
    ):  # the easiest and most common situation - just 1 optimal index
        last_el_first_contiguous_region = 0
    else:
        raise Exception("this case has not been handled")
    indices_opt = indices_opt_init[: last_el_first_contiguous_region + 1]  # [12, 13]
    return indices_opt


@jit(nopython=True)
def minimum(arr1, arr2):
    # take element-wise minimum of 2 arrays compatible with numba (instead of np.minimum(arr1, arr2))
    return arr1 * (arr1 < arr2) + arr2 * (arr1 >= arr2)


@jit(nopython=True)
def clip(val, val_min, val_max):
    # identical to np.clip
    return min(max(val, val_min), val_max)


def print_arr(arr):
    """Pretty printing of a 2D numpy array."""
    for i, row in enumerate(arr):
        string = ""
        for el in row:
            string += "{:.3f} ".format(el)
        print(i + 1, string)


def extract_hyperparam(model_name, substr):
    return model_name.split(substr)[1].split(" ")[0]


def finalize_curr_row(latex_str, weak_learner, flag_n_trees_latex):
    # finalizing the current row: apply boldfacing and add \\
    # (relies on the fact that we have only 3 metrics, i.e. TE,RTE,URTE or TE,LRTE,URTE or 4 metrics if flag_n_trees_latex is on)
    # result: 'breast-cancer & 0.3 & 0.7 & 85.4 & 85.4 & 5.1 & 11.7 & 11.7 & 5.1 & 11.7 & 11.7'
    curr_row = latex_str.split(r"\\")[-1]
    curr_str_bf = " & ".join(curr_row.split(" & ")[:2]) + " &   "
    metrics_str = " & ".join(curr_row.split(" & ")[2:])
    n_metrics = 4 if flag_n_trees_latex else 3
    metrics_curr_row = dict([(i, []) for i in range(n_metrics)])
    # result: {0: [0.7, 5.1, 5.1], 1: [85.4, 11.7, 11.7], 2: [85.4, 11.7, 11.7]}
    for i_val, val_str in enumerate(metrics_str.split(" & ")):
        # for n_trees we need int, for the rest float
        val = (
            int(val_str)
            if flag_n_trees_latex and i_val % n_metrics == n_metrics - 1
            else float(val_str)
        )
        metrics_curr_row[i_val % n_metrics].append(val)
    # form the boldfaced str that corresponds to the current row
    for tup in zip(*metrics_curr_row.values()):
        for i_m, m in enumerate(tup):
            # boldfacing condition: if minimum and it's not the number of trees (if the flag is turned on)
            if (
                m == min(metrics_curr_row[i_m])
                and not (flag_n_trees_latex and i_m == 3)
                and not (weak_learner == "stump" and i_m == 2)
            ):  # if URTE for stumps, don't boldface
                curr_str_bf += "\\textbf{" + str(m) + "} & "
            else:
                curr_str_bf += "{} & ".format(m)
        curr_str_bf += "  "  # just a margin for better latex code quality
    curr_str_bf = curr_str_bf.strip()[:-1]  # get rid of the last ' &   '
    curr_row_final = curr_str_bf + r"\\" + "\n"  # new table line
    return curr_row_final


def get_model_names(datasets, models, exp_folder, weak_learner, tree_depth):
    model_names = []
    for dataset in datasets:
        for model in models:
            depth_str = "max_depth=" + str(tree_depth) if weak_learner == "tree" else ""
            search_str = "{}/*dataset={} weak_learner={} model={}*{}*.metrics".format(
                exp_folder, dataset, weak_learner, model, depth_str
            )
            model_names_curr = glob.glob(search_str)
            model_names_curr.sort(key=lambda x: os.path.getmtime(x))
            if model_names_curr != []:
                # model_name_final = model_names_curr[-1]
                for model_name_final in model_names_curr:
                    model_name_final = model_name_final.split(".metrics")[0].split(
                        exp_folder + "/"
                    )[1]
                    model_names.append(model_name_final)
    return model_names


def get_n_proc(n_ex):
    if n_ex > 40000:
        n_proc = 50
    elif n_ex > 20000:
        n_proc = 40
    elif n_ex > 2500:
        n_proc = 25
    elif n_ex > 1000:
        n_proc = 10
    elif n_ex > 200:
        n_proc = 5
    else:
        n_proc = 1
    return n_proc



import numpy as np
from .tree_ensemble import Tree


class OneVsAllClassifier:
    def __init__(self, models):
        self.models = models
        self.n_clsf = len(models)

    def predict(self, X):
        preds = np.zeros([self.n_clsf, X.shape[0]])
        for i_cls in range(self.n_clsf):
            preds[i_cls] = self.models[i_cls].predict(X)
        return preds

    def predict_class(self, X):
        preds = self.predict(X)
        if self.n_clsf == 1:
            return (0.5 * ((preds > 0) + 1)).astype(int).flatten()
        else:
            return np.argmax(preds, 0)

    def certify_treewise(self, X, y, eps):
        preds = np.zeros([self.n_clsf, X.shape[0]])
        for i_cls in range(self.n_clsf):
            preds[i_cls] = self.models[i_cls].certify_treewise(X, y[i_cls], eps)
        return preds

    def certify_exact(self, X, y, eps):
        preds = np.zeros([self.n_clsf, X.shape[0]])
        for i_cls in range(self.n_clsf):
            preds[i_cls] = self.models[i_cls].certify_exact(X, y[i_cls], eps)
        return preds

    def fmargin(self, X, y, fx_vals=None):
        if fx_vals is None:  # if fx_vals have not been provided
            fx_vals = self.predict(X)
        if self.n_clsf > 1:
            preds_correct_class = (fx_vals * (y == 1)).sum(0, keepdims=True)
            diff = (
                preds_correct_class - fx_vals
            )  # difference between the correct class and all other classes
            diff[y == 1] = np.inf  # to exclude zeros coming from f_correct - f_correct
            fx_vals = diff.min(0, keepdims=True)
        else:
            fx_vals = y * fx_vals
        return fx_vals[0]

    def fmargin_treewise(self, X, y, eps, fx_vals=None):
        if fx_vals is None:  # if fx_vals have not been provided
            fx_vals = self.certify_treewise(X, y, eps)
        if self.n_clsf > 1:
            cert_correct_class = (fx_vals * (y == 1)).sum(0, keepdims=True)
            diff = (
                cert_correct_class + fx_vals
            )  # plus because of [min -f] in cert for all classes
            fx_vals = np.min(diff, 0, keepdims=True)
        return fx_vals[0]

    def fmargin_exact(self, X, y, eps):
        fx_vals = self.certify_exact(X, y, eps)
        if self.n_clsf > 1:
            cert_correct_class = (fx_vals * (y == 1)).sum(0, keepdims=True)
            diff = (
                cert_correct_class + fx_vals
            )  # plus because of [min -f] in cert for all classes
            fx_vals = np.min(diff, 0, keepdims=True)
        return fx_vals[0]

    def save(self, model_path):
        if model_path != "":
            model_lst = []
            for model in self.models:
                model_lst.append(model.export_model())
            model_arr = np.array(model_lst)
            np.save(model_path, model_arr)

    def load(self, model_path, iteration=-1):
        model_data = np.load(model_path, allow_pickle=True)
        for i_clsf in range(self.n_clsf):
            self.models[i_clsf].load(model_data[i_clsf], iteration)
        if type(model_data[0]) is dict:
            n_trees = max(model_data[0].keys()) + 1
        else:
            n_trees = model_data.shape[1]
        true_iteration = iteration + 1 if iteration != -1 else n_trees
        print(
            "Ensemble of {}/{} trees restored: {}".format(
                true_iteration, n_trees, model_path
            )
        )

    def dump_model(self):
        """Returns the model in JSON format compatible with XGBoost."""
        # Works for trees
        n_cls = len(self.models)
        n_trees = max([len(model.trees) for model in self.models])

        list_of_tree_dicts = []
        for i_tree in range(n_trees):
            for i_cls in range(n_cls):
                if i_tree < len(self.models[i_cls].trees):
                    tree = self.models[i_cls].trees[i_tree]
                else:
                    tree = Tree()
                tree_dict, _ = tree.get_json_dict(counter_terminal_nodes=-10)
                list_of_tree_dicts.append(tree_dict)

        return list_of_tree_dicts



import numpy as np
from numba import jit
from collections import OrderedDict
from .robust_boosting import (
    exp_loss_robust,
    dtype,
    fit_plain_stumps,
    fit_robust_bound_stumps,
    fit_robust_exact_stumps,
)
from .utils import (
    minimum,
    get_contiguous_indices,
    get_n_proc,
)
from concurrent.futures import ThreadPoolExecutor


class Stump:
    def __init__(self, w_l, w_r, b, coord, loss):
        # `loss` is the loss of the whole ensemble after applying this stump
        self.w_l, self.w_r, self.b, self.coord, self.loss = w_l, w_r, b, coord, loss
        self.left, self.right = None, None

    def predict(self, X):
        value = self.w_l + self.w_r * (X[:, self.coord] >= self.b)
        return value

    def find_min_yf(self, X, y, eps):
        split_lbs, split_ubs = X[:, self.coord] - eps, X[:, self.coord] + eps
        lval, rval = self.w_l, self.w_r + self.w_l

        # Fast vectorized version
        guaranteed_left = split_ubs < self.b
        guaranteed_right = split_lbs > self.b
        uncertain = (split_lbs <= self.b) * (split_ubs >= self.b)

        lbs = (
            y * lval * guaranteed_left
            + y * rval * guaranteed_right
            + np.minimum(y * lval, y * rval) * uncertain
        )

        return lbs

    def __repr__(self):
        lval, rval, threshold = self.w_l, self.w_r + self.w_l, self.b
        return "Tree: if x[{}] < {:.4f}: {:.4f} else {:.4f}".format(
            self.coord, threshold, lval, rval
        )

    def get_json_dict(self, counter_terminal_nodes):
        """
        counter_terminal_nodes: not used here
        """
        precision = 5
        children_list = [
            {"nodeid": 1, "leaf": round(self.w_l, precision)},
            {"nodeid": 2, "leaf": round(self.w_l + self.w_r, precision)},
        ]
        stump_dict = {
            "nodeid": 0,
            "split": "f" + str(int(self.coord)),
            "split_condition": round(self.b, precision),
            "yes": 1,
            "no": 2,
            "children": children_list,
        }

        return stump_dict, counter_terminal_nodes


class StumpEnsemble:
    def __init__(
        self, weak_learner, n_trials_coord, lr, idx_clsf, n_bins=-1, max_weight=1.0
    ):
        self.weak_learner = weak_learner
        self.n_trials_coord = n_trials_coord
        self.lr = lr
        self.idx_clsf = idx_clsf
        self.n_bins = n_bins
        self.max_weight = max_weight
        self.trees = []
        self.coords_trees = OrderedDict()

    def __repr__(self):
        sorted_trees = sorted(self.trees, key=lambda tree: tree.coord)
        return "\n".join([str(t) for t in sorted_trees])

    def copy(self):
        ensemble_new = StumpEnsemble(
            self.weak_learner,
            self.n_trials_coord,
            self.lr,
            self.idx_clsf,
            self.n_bins,
            self.max_weight,
        )
        for tree in self.trees:
            ensemble_new.add_weak_learner(tree, apply_lr=False)
        return ensemble_new

    def load(self, ensemble_arr, iteration=-1):
        if iteration != -1:  # take up to some iteration
            ensemble_arr = ensemble_arr[: iteration + 1]
        for i in range(ensemble_arr.shape[0]):
            w_l, w_r, b, coord, loss = ensemble_arr[i, :]
            coord = int(coord)
            tree = Stump(w_l, w_r, b, coord, loss)
            # the values of w_l and w_r should be already scaled by lr, would be wrong to do this again
            self.add_weak_learner(tree, apply_lr=False)

    def export_model(self):
        ensemble_arr = np.zeros([len(self.trees), 5])
        for i, tree in enumerate(self.trees):
            ensemble_arr[i, :] = [tree.w_l, tree.w_r, tree.b, tree.coord, tree.loss]
        return ensemble_arr

    def save(self, path):
        if path != "":
            np.save(path, self.export_model(), allow_pickle=False)

    def add_weak_learner(self, tree, apply_lr=True):
        if apply_lr:
            tree.w_l, tree.w_r = tree.w_l * self.lr, tree.w_r * self.lr
        self.trees.append(tree)
        if tree.coord not in self.coords_trees:
            self.coords_trees[tree.coord] = []
        self.coords_trees[tree.coord].append(tree)

    def add_empty_weak_learner(self):
        empty_stump = Stump(0.0, 0.0, 0.0, 0, 0.0)
        self.add_weak_learner(empty_stump)

    def predict(self, X):
        Fx = np.zeros(X.shape[0])
        for tree in self.trees:
            Fx += tree.predict(X)
        return Fx

    def attack_by_sampling(self, X, y, eps, n_trials):
        """A simple attack just by sampling in the Linf-box around the points. More of a sanity check."""
        num, dim = X.shape
        f_x_vals = np.zeros((num, n_trials))
        # Note: for efficiency, we sample the same random direction for all points
        deltas = np.random.uniform(-eps, eps, size=(dim, n_trials))
        for i in range(n_trials - 1):
            # let's keep them as real images, although not strictly needed
            perturbed_pts = np.clip(X + deltas[:, i], 0.0, 1.0)
            f_x_vals[:, i] = self.predict(perturbed_pts)
        # maybe in some corner cases, the predictions at the original point is more worst-case than the sampled points
        f_x_vals[:, n_trials - 1] = self.predict(X)

        f_x_min = np.min(y[:, None] * f_x_vals, axis=1)
        return f_x_min

    def certify_treewise(self, X, y, eps):
        lb_ensemble = np.zeros(X.shape[0])

        # The naive tree-wise bounded on the merged trees
        for tree in self.trees:
            lb_ensemble += tree.find_min_yf(X, y, eps)
        return lb_ensemble

    @staticmethod
    @jit(nopython=True)
    def find_min_coord_diff(X_proj, y, thresholds, w_r_values, eps):
        # parallel=True doesn't help here; not sure if jit here is helpful at all. maybe if there are many thresholds
        num = X_proj.shape[0]
        idx = np.argsort(thresholds)
        sorted_thresholds = thresholds[idx]
        sorted_w_r = w_r_values[idx]
        f_x_min_coord_diff, f_x_cumsum = np.zeros(num), np.zeros(num)
        for i_t in range(len(sorted_thresholds)):
            # consider the threshold if it belongs to (x-eps, x+eps] (x-eps is excluded since already evaluated)
            idx_x_eps_close_to_threshold = (X_proj - eps < sorted_thresholds[i_t]) * (
                sorted_thresholds[i_t] <= X_proj + eps
            )
            f_diff = y * sorted_w_r[i_t] * idx_x_eps_close_to_threshold
            f_x_cumsum += f_diff
            f_x_min_coord_diff = minimum(f_x_cumsum, f_x_min_coord_diff)
        return f_x_min_coord_diff

    def certify_exact(self, X, y, eps, coords_to_ignore=()):
        # Idea: iterate over all thresholds, and then check if they are in (x-eps, x+eps]
        num, dim = X.shape
        f_x_min = np.zeros(num)

        # Fast, vectorized version
        for coord in self.coords_trees.keys():
            if coord in coords_to_ignore:
                continue
            trees_current_coord = self.coords_trees[coord]

            f_x_min_coord_base = np.zeros(num)
            thresholds, w_r_values = (
                np.zeros(len(trees_current_coord)),
                np.zeros(len(trees_current_coord)),
            )
            for i in range(len(trees_current_coord)):
                tree = trees_current_coord[i]
                f_x_min_coord_base += y * tree.predict(X - eps)
                thresholds[i], w_r_values[i] = tree.b, tree.w_r

            # merge trees with the same thresholds to prevent an overestimation (lower bounding) of the true minimum
            thresholds_list, w_r_values_list = [], []
            for threshold in np.unique(thresholds):
                thresholds_list.append(threshold)
                w_r_values_list.append(np.sum(w_r_values[thresholds == threshold]))
            thresholds, w_r_values = (
                np.array(thresholds_list),
                np.array(w_r_values_list),
            )

            f_x_min += f_x_min_coord_base + self.find_min_coord_diff(
                X[:, coord], y, thresholds, w_r_values, eps
            )
        return f_x_min

    def fit_stumps_over_coords(self, X, y, gamma, model, eps):
        verbose = False
        parallel = False  # can speed up the training on large datasets
        n_ex = X.shape[0]
        X, y, gamma = X.astype(dtype), y.astype(dtype), gamma.astype(dtype)
        prev_loss = np.mean(gamma)

        # 151 features are always 0.0 on MNIST 2 vs 6. And this number is even higher for smaller subsets of MNIST,
        # i.e. subsets of examples partitioned by tree splits.
        idx_non_trivial = np.abs(X).sum(axis=0) > 0.0
        features_to_check = np.random.permutation(np.where(idx_non_trivial)[0])[
            : self.n_trials_coord
        ]

        n_coords = len(features_to_check)
        params, min_losses = np.zeros((n_coords, 4)), np.full(n_coords, np.inf)

        if parallel:
            n_proc = get_n_proc(n_ex)
            n_proc = min(n_coords, min(100, n_proc))
            batch_size = n_coords // n_proc
            n_batches = n_coords // batch_size + 1

            with ThreadPoolExecutor(max_workers=n_proc) as executor:
                procs = []
                for i_batch in range(n_batches):
                    coords = features_to_check[
                        i_batch * batch_size : (i_batch + 1) * batch_size
                    ]
                    args = (X, X[:, coords], y, gamma, model, eps, coords)
                    procs.append(executor.submit(self.fit_stump_batch, *args))

                # Process the results
                i_coord = 0
                for i_batch in range(n_batches):
                    res_many = procs[i_batch].result()
                    for res in res_many:
                        min_losses[i_coord], *params[i_coord, :] = res
                        i_coord += 1
        else:
            for i_coord, coord in enumerate(features_to_check):
                min_losses[i_coord], *params[i_coord, :] = self.fit_stump(
                    X, X[:, coord], y, gamma, model, eps, coord
                )

        id_best_coord = min_losses.argmin()
        min_loss = min_losses[id_best_coord]
        best_coord = int(
            params[id_best_coord][3]
        )  # float to int is necessary for a coordinate
        best_wl, best_wr, best_b = (
            params[id_best_coord][0],
            params[id_best_coord][1],
            np.float32(params[id_best_coord][2]),
        )
        if verbose:
            print(
                "[{}-vs-all]: n_ex {}, n_coords {} -- loss {:.5f}->{:.5f}, b={:.3f} wl={:.3f} wr={:.3f} at coord {}".format(
                    self.idx_clsf,
                    n_ex,
                    n_coords,
                    prev_loss,
                    min_loss,
                    best_b,
                    best_wl,
                    best_wr,
                    best_coord,
                )
            )
        return Stump(best_wl, best_wr, best_b, best_coord, min_loss)

    def fit_stump_batch(self, X, Xs, y, gamma, model, eps, coords):
        res = np.zeros([len(coords), 5])
        for i, coord in enumerate(coords):
            res[i] = self.fit_stump(X, Xs[:, i], y, gamma, model, eps, coord)
        return res

    def fit_stump(self, X, X_proj, y, gamma_global, model, eps, coord):
        min_prec_val = 1e-7
        min_val, max_val = (
            0.0,
            1.0,
        )  # can be changed if the features are in a different range
        n_bins = self.n_bins

        # Needed for exact robust optimization with stumps
        trees_current_coord = (
            self.coords_trees[coord] if coord in self.coords_trees else []
        )
        w_rs, bs = (
            np.zeros(len(trees_current_coord)),
            np.zeros(len(trees_current_coord)),
        )
        for i in range(len(trees_current_coord)):
            w_rs[i] = trees_current_coord[i].w_r
            bs[i] = trees_current_coord[i].b

        if (
            model == "robust_exact" and trees_current_coord != []
        ):  # note: the previous gamma is just ignored
            min_Fx_y_exact_without_j = self.certify_exact(
                X, y, eps, coords_to_ignore=(coord,)
            )
            w_ls = np.sum([tree.w_l for tree in trees_current_coord])
            gamma = np.exp(-min_Fx_y_exact_without_j - y * w_ls)
        else:
            gamma = gamma_global

        if n_bins > 0:
            if model == "robust_bound":
                # b_vals = np.array([0.31, 0.41, 0.5, 0.59, 0.69])  # that's the thresholds that one gets with n_bins=10
                b_vals = np.arange(eps * n_bins, n_bins - eps * n_bins + 1) / n_bins
                # to have some margin to make the thresholds not adversarially reachable from 0 or 1
                b_vals[b_vals < 0.5] += 0.1 * 1 / n_bins
                b_vals[b_vals > 0.5] -= 0.1 * 1 / n_bins
            else:
                b_vals = np.arange(1, n_bins) / n_bins
        else:
            threshold_candidates = np.sort(X_proj)
            if (
                len(threshold_candidates) == 0
            ):  # if no samples left according to min_samples_leaf
                return [np.inf, 0.0, 0.0, 0.0, -1]
            if (
                model not in ["robust_bound", "robust_exact"] or eps == 0.0
            ):  # plain, da_uniform or at_cube training
                b_vals = np.copy(threshold_candidates)
                b_vals += min_prec_val  # to break the ties
            else:  # robust training
                b_vals = np.concatenate(
                    (threshold_candidates - eps, threshold_candidates + eps), axis=0
                )
                b_vals = np.clip(
                    b_vals, min_val, max_val
                )  # save computations (often goes 512 -> 360 thresholds on MNIST)
                # to make in the overlapping case [---x-[--]-x---] output 2 different losses in the middle
                n_bs = len(threshold_candidates)
                b_vals += np.concatenate(
                    (-np.full(n_bs, min_prec_val), np.full(n_bs, min_prec_val)), axis=0
                )
            b_vals = np.unique(b_vals)  # use only unique b's
            b_vals = np.sort(
                b_vals
            )  # still important to sort because of the final threshold selection

        if model in ["plain", "da_uniform", "at_cube"]:
            losses, w_l_vals, w_r_vals, b_vals = fit_plain_stumps(
                X_proj, y, gamma, b_vals, self.max_weight
            )
        elif model == "robust_bound":
            losses, w_l_vals, w_r_vals, b_vals = fit_robust_bound_stumps(
                X_proj, y, gamma, b_vals, eps, self.max_weight
            )
        elif model == "robust_exact":
            losses, w_l_vals, w_r_vals, b_vals = fit_robust_exact_stumps(
                X_proj, y, gamma, b_vals, eps, w_rs, bs, self.max_weight
            )
        else:
            raise ValueError("wrong model")

        min_loss = np.min(losses)
        # probably, they are already sorted, but to be 100% sure since it is not explicitly mentioned in the docs
        indices_opt_init = np.sort(np.where(losses == min_loss)[0])
        indices_opt = get_contiguous_indices(indices_opt_init)
        id_opt = indices_opt[len(indices_opt) // 2]

        idx_prev = np.clip(
            indices_opt[0] - 1, 0, len(b_vals) - 1
        )  # to prevent stepping out of the array
        idx_next = np.clip(
            indices_opt[-1] + 1, 0, len(b_vals) - 1
        )  # to prevent stepping out of the array
        b_prev, w_l_prev, w_r_prev = (
            b_vals[idx_prev],
            w_l_vals[idx_prev],
            w_r_vals[idx_prev],
        )
        b_next, w_l_next, w_r_next = (
            b_vals[idx_next],
            w_l_vals[idx_next],
            w_r_vals[idx_next],
        )
        # initialization
        b_leftmost, b_rightmost = b_vals[indices_opt[0]], b_vals[indices_opt[-1]]
        # more involved, since with +-eps, an additional check of the loss is needed
        if model in ["plain", "da_uniform", "at_cube"]:
            b_rightmost = b_next
        elif model in ["robust_bound", "robust_exact"]:
            h_flag = False if model == "robust_bound" else True

            b_prev_half = (b_prev + b_vals[indices_opt[0]]) / 2
            loss_prev_half = exp_loss_robust(
                X_proj, y, gamma, w_l_prev, w_r_prev, w_rs, bs, b_prev_half, eps, h_flag
            )

            b_next_half = (b_vals[indices_opt[-1]] + b_next) / 2
            loss_next_half = exp_loss_robust(
                X_proj, y, gamma, w_l_next, w_r_next, w_rs, bs, b_next_half, eps, h_flag
            )

            # we extend the interval of the constant loss to the left and to the right if there the loss is
            # the same at b_prev_half or b_next_half
            if loss_prev_half == losses[id_opt]:
                b_leftmost = b_prev
            if loss_next_half == losses[id_opt]:
                b_rightmost = b_next
        else:
            raise ValueError("wrong model")

        # we put in the middle of the interval of the constant loss
        b_opt = (b_leftmost + b_rightmost) / 2

        # For the chosen threshold, we need to calculate w_l, w_r
        # Some of w_l, w_r that correspond to min_loss may not be optimal anymore
        b_val_final = np.array([b_opt])
        if model in ["plain", "da_uniform", "at_cube"]:
            loss, w_l_opt, w_r_opt, _ = fit_plain_stumps(
                X_proj, y, gamma, b_val_final, self.max_weight
            )
        elif model == "robust_bound":
            loss, w_l_opt, w_r_opt, _ = fit_robust_bound_stumps(
                X_proj, y, gamma, b_val_final, eps, self.max_weight
            )
        elif model == "robust_exact":
            loss, w_l_opt, w_r_opt, _ = fit_robust_exact_stumps(
                X_proj, y, gamma, b_val_final, eps, w_rs, bs, self.max_weight
            )
        else:
            raise ValueError("wrong model")
        loss, w_l_opt, w_r_opt = loss[0], w_l_opt[0], w_r_opt[0]
        # recalculation of w_l, w_r shouldn't change the min loss

        if np.abs(loss - min_loss) > 1e7:
            print("New loss: {:.5f}, min loss before: {:.5f}".format(loss, min_loss))

        best_loss = losses[id_opt]
        return [best_loss, w_l_opt, w_r_opt, b_opt, coord]



from ..attack import AttackWrapper

import numpy as np

from tqdm import tqdm

from collections import defaultdict

from copy import deepcopy


def _extract_bounding_boxes(tree, bounds):
    if "leaf" in tree:
        return [(deepcopy(bounds), tree["leaf"])]
    else:
        leaves = []

        # If the split's new right threshold (so the node on the left) is more specific
        # than the previous one, update the bound and recurse
        old_bound = bounds[tree["split"]][1]
        if (
            tree["split_condition"] <= bounds[tree["split"]][1]
            and tree["split_condition"] >= bounds[tree["split"]][0]
        ):
            bounds[tree["split"]][1] = tree["split_condition"]

        if tree["split_condition"] >= bounds[tree["split"]][0]:
            for subtree in tree["children"]:
                if subtree["nodeid"] == tree["yes"]:
                    leaves.extend(_extract_bounding_boxes(subtree, bounds))

        bounds[tree["split"]][1] = old_bound

        # If the split's new left threshold (so the node on the right) is more specific
        # than the previous one, update the bound and recurse
        old_bound = bounds[tree["split"]][0]
        if (
            tree["split_condition"] >= bounds[tree["split"]][0]
            and tree["split_condition"] <= bounds[tree["split"]][1]
        ):
            bounds[tree["split"]][0] = tree["split_condition"]

        if tree["split_condition"] < bounds[tree["split"]][1]:
            for subtree in tree["children"]:
                if subtree["nodeid"] == tree["no"]:
                    leaves.extend(_extract_bounding_boxes(subtree, bounds))

        bounds[tree["split"]][0] = old_bound

        return leaves


class DecisionTreeAttackWrapper(AttackWrapper):
    def __init__(self, json_model, n_classes):
        if len(json_model) != 1:
            raise ValueError("This attack can only be used with single decision trees")
        self.json_model = json_model

        if n_classes != 2:
            raise ValueError("Currently only binary classification is supported")
        self.n_classes = n_classes

        bounds = defaultdict(lambda: np.array([-np.inf, np.inf]))
        self.leaves = _extract_bounding_boxes(json_model[0], bounds)

    def adversarial_examples(self, X, y, order, options={}):
        # Turn 'leaves' into bounding boxes and leaf prediction values
        bound_dicts, leaf_values = zip(*self.leaves)
        predictions = [value > 0 for value in leaf_values]
        bounding_boxes = []
        for bound_dict in bound_dicts:
            bounding_box = np.tile(np.array([-np.inf, np.inf]), (X.shape[1], 1))
            for i, bound in bound_dict.items():
                bounding_box[i] = bound
            bounding_boxes.append(bounding_box)

        X_adv = []
        for sample, label in tqdm(zip(X, y), total=X.shape[0]):
            # Create a minimal adversarial example for each leaf
            # then choose the one with minimal distance
            best_distance = np.inf
            for bounding_box, prediction in zip(bounding_boxes, predictions):
                if prediction != label:
                    adv_example = np.clip(
                        sample, bounding_box[:, 0], bounding_box[:, 1]
                    )
                    distance = np.linalg.norm(adv_example - sample, ord=order)

                    if distance < best_distance:
                        best_distance = distance
                        best_adv_example = adv_example

            if best_distance == np.inf:
                # If no adversarial example is possible (tree predicts one value everywhere)
                # then return a vector of NaNs
                X_adv.append(np.full(len(sample), np.nan))
            else:
                X_adv.append(best_adv_example)

        return np.array(X_adv)
